import plotly.graph_objects as go
import plotly.express as px
import numpy as np
import warnings
warnings.filterwarnings('ignore')
from plotly.subplots import make_subplots

rows = 1
vertical_spacing=0.2/rows
colorbar_= True

def subplots2D(image,height=400,width=900,sharex=False,sharey=False,x_title=None,y_title=None,
               colorbar=True,title='',layout={},col_w=None,cols=2):
    
    global rows, vertical_spacing, colorbar_
    colorbar_ = colorbar

    num = len(image)
    rows = num // cols if num % cols == 0 else num // cols + 1
    vertical_spacing=0.2/rows
    height = height if rows<2 else 200
    cols, column_widths = (cols * 2,[0.45,0.05]*cols) if colorbar else (cols,[0.5]*cols)
    column_widths = column_widths if col_w is None else col_w
    fig = go.Figure(layout=go.Layout(height=height*rows, width=width))
    fig = make_subplots(rows=rows,cols=cols,shared_xaxes=sharex,
                        shared_yaxes=sharey,
                       column_widths=column_widths,
                       row_heights=[1/rows]*rows,
                       vertical_spacing=vertical_spacing,
                       horizontal_spacing=0.04,
                       x_title=x_title,
                       y_title=y_title,
                       figure=fig)
    for i,trace in enumerate(image):
        if colorbar:
            r, c = (i//cols+1,2*i%cols+1) #if i%2==0 else (i//cols+1,3)
        else:
            r, c = (i//2+1,i%cols+1) #if i%2==0 else (i//2+1,2)
        
        for tr in trace:
            fig.add_trace(tr,row=r,col=c)
            try:
                yanchor = 1-(i//2*2+1)/2/rows*(1-vertical_spacing*(rows-1)) - vertical_spacing*(i//2)
                x_bar, y_bar = (0.39,yanchor) if i%2==0 else (0.91,yanchor)
                name = tr.name
                if name[:2] == 'cb':
                    fig.update_traces(colorbar=dict(x=x_bar,y=y_bar,len=0.8/rows,title=name[2:],
                                                    nticks=5),row=r,col=c)
            except:
                pass
            
    legend=dict(
#         x=1,
#         y=1,
        # traceorder="reversed",
        title_font_family="Times New Roman",
        font=dict(
            family="Courier",
            size=12,
            color="black"
        ),
#         bgcolor="LightSteelBlue",
#         bordercolor="Black",
        borderwidth=1
    )
    fig.update_layout(margin=dict(r=0,t=30,b=55,l=80),showlegend=True,legend=legend,title=title,**layout)#,xaxis=dict(fixedrange=True),yaxis=dict(fixedrange=True))

    return fig

def add_plot(fig,tr,row=0,col=0):
    global rows, vertical_spacing, colorbar_
    if colorbar_:
        r, c = (row,2*col+1)#(row,1) if col==1 else (row,3)
    else:
        r, c = (row,col)#(row,1) if col==1 else (row,2)
    fig.add_trace(tr,row=r,col=c)
    try:
        name = tr.name
        yanchor = 1-(r*2-1)/2/rows*(1-vertical_spacing*(rows-1)) - vertical_spacing*(r-1)
        x_bar, y_bar = (0.39,yanchor) if c==1 else (0.91,yanchor)
        if name[:2] == 'cb':
            fig.update_traces(colorbar=dict(x=x_bar,y=y_bar,len=0.8/rows,title=name[2:],
                                                            nticks=5),row=r,col=c)
    except:
        pass
    
def add_title(fig,title):
    name = fig.layout.title.text
    name += f'{title}'
    fig.update_layout(title=name)
    
def set_title(fig,title):
    name = f'{title}'
    fig.update_layout(title=name)


import numpy as np
import scipy as sp 
import sympy as sy
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from . import optimize as op
import math, time as TM
import pprint
from itertools import product


xvar = sy.Symbol('x',real=True)

def func(paras,b0,bias=None):
    voffset, vperiod, ejs, ec, d = paras
    sign = 1 if b0>=voffset else -1
    x = np.linspace(-vperiod/2,vperiod/2,1001) + voffset if bias is None else bias
    tmp = np.pi*(x-voffset)/vperiod
    f01 = np.sqrt(8*ejs*ec) * (np.abs(np.cos(tmp)) * np.sqrt(1+d**2*np.tan(tmp)**2) - 1) * (sign*np.sign(x-voffset)) + np.sqrt(8*ejs*ec) -ec
    return x, f01

def freq2flux(delta,f0,b0,paras,numofinterp=1001):
    
    bias, f01 = func(paras,b0)
    _, funcp = op.RowToRipe().interp1d(f01,bias)
    return funcp(f0+delta)

def Print(*data,width=30, compact=True, indent=1, depth=3):
    
    pp = pprint.PrettyPrinter(width=width, compact=compact, 
                              indent=indent, depth=depth)
    pp.pprint(data)
    
def printColor(x,mode=1,color_b='w',color_f='r'):
    clst_b = {'w':47,'k':40,'r':41,'g':42,'y':43,'b':44}
    clst_f = {'w':37,'k':30,'r':31,'g':32,'y':33,'b':34}
    print(f'\033[{mode};{clst_f[color_f]};{clst_b[color_b]}m{x}\033[0m')

def printTime(FORMAT="%Y/%m/%d %A %H:%M:%S",printinfo=True):
    x = TM.strftime(FORMAT,TM.localtime())
    if printinfo:
        print(x)
    return x

def bitstring(level=2,repeat=1,dtype='str'):
    
    if dtype == 'str':
        x = product([str(i) for i in range(level)],repeat=repeat)
        return iter([''.join(j) for j in x])
    else:
        x = product(range(level),repeat=repeat)
        return x
    
def bitstring_in_constrain(n=10,k=5,dtype='str'):
    from itertools import combinations
    if dtype == 'str':
        bstrings = [format(sum(1 << i for i in comb), f'0{n}b') for comb in combinations(range(n), k)]
    if dtype == 'tuple':
        bstrings = [tuple(int(b) for b in format(sum(1 << i for i in comb), f'0{n}b')) for comb in combinations(range(n), k)]
    return bstrings
    
def numTobinstring(num,ndigits=None):
    if ndigits is None:
        return format(num,'b')
    else:
        x = format(num,'b')
        l = len(x)
        if l != ndigits:
            x = '0'*(ndigits-len(x)) + x
        return x
    
def firstgreater(y,target,x=None):
    index = np.argmax(y>=target)
    if x is None:
        return index, None
    else:
        x = np.array(x)
        return index, x[index]
    
def choose_min_or_max(data):
    index = (data!=np.min(data))*(data!=np.max(data))
    median = np.median(data[index])
    l_min = np.max(np.abs(np.min(data)-median))
    l_max = np.max(np.abs(np.max(data)-median))
    if l_min >= l_max:
        return np.argmin(data)
    else:
        return np.argmax(data)

def nearest(y,target,x=None):
    y = np.array(y)
    if x is None:
        index = np.argmin(np.abs(y-target))
        return index, None
    else:
        x = np.array(x)
        index = np.argmin(np.abs(y-target))
        return index, x[index]

def maxFreq(paras):
    """
    fc: qubit's frequency at known point
    bias=bias(f_target) - bias(fc)
    """
    func,voffset, vperiod = op.Spec2d_Fit().func_f01, paras[0], paras[1]
    v = np.linspace(-vperiod/2,vperiod/2,50001) + voffset
    y = func
    fmax = np.max(y(v,paras))
    return fmax
# def specfunc2data(func,)
def specshift(paras,fc,bias=0,side='lower'):
    """
    fc: qubit's frequency at known point
    bias=bias(f_target) - bias(fc)
    """
    func,voffset, vperiod = op.Spec2d_Fit().func_f01, paras[0], paras[1]
    v = np.linspace(-vperiod/2,vperiod/2,50001) + voffset
    # y = sy.lambdify(xvar,func,'numpy')
    y = func
    if side == 'lower':
        f01 = y(v,paras)[v<voffset]
        _, vtarget = nearest(f01,fc,v[v<voffset])
        vdiff = voffset - vtarget
        paras[0] += vdiff
        # v = np.linspace(-vperiod/2,vperiod/2,50001) + (voffset+vdiff)
        return y(bias,paras), (vtarget+bias)

    if side == 'higher':
        f01 = y(v,paras)[v>voffset]
        _, vtarget = nearest(f01,fc,v[v>voffset])
        vdiff = voffset - vtarget
        paras[0] += vdiff
        # v = np.linspace(-vperiod/2,vperiod/2,50001) + (voffset+vdiff)
        return y(bias,paras), (vtarget+bias)

def biasshift(paras,fc,fshift=0,side='lower'):
    """
    fc: qubit's frequency at known point
    fshift: = f_target - fc
    """
    func,voffset, vperiod = op.Spec2d_Fit().func_f01, paras[0], paras[1]
    v = np.linspace(-vperiod/2,vperiod/2,50001) + voffset
    # y = sy.lambdify(xvar,func,'numpy')
    y = func
    if np.max(fc+fshift)>np.max(y(v,paras)):
        raise('too big')
    if side == 'lower':
        vnew = v[v<voffset]
        f01 = y(v,paras)[v<voffset]
        finterp = sp.interpolate.interp1d(f01,vnew)
        _, vtarget = nearest(f01,fc,vnew)
        return finterp(fc+fshift)-vtarget, vtarget
    if side == 'higher':
        vnew = v[v>voffset]
        f01 = y(v,paras)[v>voffset]
        finterp = sp.interpolate.interp1d(f01,vnew)
        _, vtarget = nearest(f01,fc,vnew)
        return finterp(fc+fshift)-vtarget, vtarget


def biasshift_res(qubit,result_fc,result_spec,fshift):
    paras = list((result_spec['meta']['config'][qubit]['specparams']).values())
    fc = result_fc['meta']['config']['gate']['R'][qubit]['params']['frequency']/1e9
    b_old = result_fc['meta']['config'][qubit]['biaslst']['idle_bias']

    b0_0 = list((result_spec['meta']['config'][qubit]['s21params']).values())[0]
    side = 'lower' if b_old<=b0_0 else 'higher'

    biasshift_C,_= biasshift(paras,fc,fshift=fshift,side=side)
    return biasshift_C


# def dresscali(funclist,dressenergy,fc,imAmp=0,side='lower'):

#     fnew = dressenergy(imAmp)-f_ex/1e9
#     bias_offset = dt.biasshift(specfuncz,f_ex/1e9,fnew,'lower')

def Z_scale_predict(q_ex,q_z,measure,f_ref=4.437e9,freq_up=120e6,freq_down=20e6):
    qubits = list(measure.qubits.values())
    up_scales = {}
    down_scales = {}
    n_bar = (q_ex.index+q_z.index)/2
    i_counts = 0
    for qubit in qubits:

        if abs(qubit.index-n_bar)<0.6:
            up_scales[qubit.q_name]=qubit.volt

        if 0.6<abs(qubit.index-n_bar)<1.6:
            f_target = f_ref+(-1)**i_counts*freq_up
            zc_ref = qubit.volt
            up_scales[qubit.q_name]=f_ref_To_f_target(qubit,f_target/1e9,f_ref/1e9,0,zc_ref=zc_ref)[1]
            i_counts += 1

        if abs(qubit.index-n_bar)>1.6:
            f_target = f_ref+(-1)**i_counts*freq_down
            zc_ref = qubit.volt
            up_scales[qubit.q_name]=f_ref_To_f_target(qubit,f_target/1e9,f_ref/1e9,0,zc_ref=zc_ref)[1] 
            i_counts += 1

    up_scales_copy = up_scales.copy()
    
    for i,qubit in enumerate(qubits):
        down_scales[qubit.q_name] = 2*qubit.volt - up_scales_copy[qubit.q_name]
    return [up_scales,down_scales]

def f_ref_To_f_target(qubit_ex,f_target,f_ref,dc_ref=0,zc_ref=0):
    volt_ex1, vtarget = biasshift(qubit_ex.specfuncz,f_ref,f_target-f_ref,side=qubit_ex.side)
    dc = volt_ex1*qubit_ex.T_bias[0]/qubit_ex.T_z[0] + dc_ref
    zc = volt_ex1 + zc_ref
    return dc, zc

def vTophi(funclist,T_bias,fc,side='lower'):
    assert fc < 10
    func,voffset, vperiod, ejs, ec, d = funclist
    vperiod, voffset = T_bias
    tmp_s = np.pi*(xvar-voffset)/vperiod
    func = sy.sqrt(8*ejs*ec*sy.Abs(sy.cos(tmp_s))*sy.sqrt(1+d**2*sy.tan(tmp_s)**2))-ec
    v = np.linspace(-vperiod/2,vperiod/2,50001) + voffset
    y = sy.lambdify(xvar,func,'numpy')
    if side == 'lower':
        vnew = v[v<voffset]
        f01 = y(v)[v<voffset]
        # finterp = sp.interpolate.interp1d(f01,vnew)
        index, vtarget = nearest(f01,fc,vnew)
        return vtarget
    if side == 'higher':
        vnew = v[v>voffset]
        f01 = y(v)[v>voffset]
        # finterp = sp.interpolate.interp1d(f01,vnew)
        index, vtarget = nearest(f01,fc,vnew)
        return vtarget

def search_idle_point(measure,f_now,f_targets,dcstate):
    print(dcstate)
    for i,j in enumerate(measure.qubitToread):
        qubit_ex=measure.qubits[j]
        f_target = f_targets[i]
        num = measure.qubitToread.index(qubit_ex.q_name)
        f_ref = f_now[num]
        dc_ref = dcstate[qubit_ex.q_name]
        dcstate[qubit_ex.q_name] = round(f_ref_To_f_target(qubit_ex,f_target,f_ref,dc_ref)[0],5)
    print(dcstate)
    return dcstate

def chi_cq(qubit):
    delta_cq = qubit.f_lo - qubit.f_ex
    g_cq, alpha = qubit.g_cq, qubit.alpha
    chi = g_cq**2/delta_cq/(1+delta_cq/alpha)
    return chi

def acstarkShift(qubit,readamp=0,side='lower'):
    res, _, chi = qubit.acstark_shift
    chiRatio = chi_cq(qubit) / chi
    freqdownShift = chiRatio*res[0]*readamp**2   ###### -2*chi*n = -2*chi*m*readamp**2
    print(freqdownShift)
    readvolt, _ = biasshift(qubit.specfuncz,qubit.f_ex/1e9,fshift=freqdownShift,side=side)
    return -readvolt

def anglehist(value,bins=100,circle=False):
#     comstd = lambda x: np.sqrt(np.std(np.real(x))**2+np.std(np.imag(x))**2)
    if circle:
        center = np.mean(value)
        r = np.std(value)
        mask = np.abs(value-center) <= r
        value = value[mask]
    comspace = lambda x: np.abs(x-np.mean(x))
    angle = np.angle(value-np.mean(value))
    space = comspace(value)
    step = math.gcd(len(angle),bins)
    index = np.argsort(angle)
    # x = angle[index]
    y = space[index]
    noise = []
    for i in range(0,len(angle)//step):
        start, end = i*step, (i+1)*step
        noise.append(np.mean(y[start:end]))
    noise_mean, noise_std = np.mean(noise), np.std(noise)
    return np.mean(value), noise_mean, noise_std, noise

# def Classify_fit(data,qnum=0,nstate=2):
#     data = np.array(data)
#     S = np.array([data[i,:,qnum] for i in range(nstate)]).flatten()
#     x,z = np.real(S), np.imag(S)
#     d = list(zip(x,z))
#     center = [(np.real(np.mean(data[i,:,qnum])),np.imag(np.mean(data[i,:,qnum]))) for i in range(nstate)] 

#     kmeans = KMeans(n_clusters=nstate,max_iter=1000,tol=0.0001,init =np.array(center))
#     kmeans.fit(d)
# #     c0, c1, c2 = kmeans.cluster_centers_[0,:], kmeans.cluster_centers_[1,:], kmeans.cluster_centers_[2,:]
#     color = kmeans.predict(d)
    
#     return x,z, kmeans.predict, kmeans.cluster_centers_, color

def Classify_fit(data,qnum=0,nstate=2):
    data = np.array(data)
    S = np.array([data[i,:,qnum] for i in range(nstate)]).flatten()
    x,z = np.real(S), np.imag(S)
    d = list(zip(x,z))
    center = [(np.real(np.mean(data[i,:,qnum])),np.imag(np.mean(data[i,:,qnum]))) for i in range(nstate)]

    kmeans = KMeans(n_clusters=nstate,max_iter=1000,tol=0.0001,init =np.array(center),n_init=1)
    kmeans.fit(d)
    center_fit = [kmeans.cluster_centers_[i,:] for i in range(nstate)]
#     c0, c1, c2 = kmeans.cluster_centers_[0,:], kmeans.cluster_centers_[1,:], kmeans.cluster_centers_[2,:]
    color = kmeans.predict(d)
    color0 = [kmeans.predict(d[i*(len(d)//nstate):(i+1)*(len(d)//nstate)]) for i in range(nstate)]
    # print(center_fit)
    p = [list(color0[i]).count(i)/len(color0[i]) for i in range(nstate)]
    
    return x,z, kmeans.predict, center_fit, color, p

def FFCC(FF,CC,f_rabi,dcstate,f_target):
    for i,q in enumerate(dcstate):
        f_t = f_target[i]
        c = CC[i]+[dcstate[q]]
        f = FF[i]+ [f_rabi[i]]
        if len(f)==3:
            index = np.argmax(abs(np.array(f)-f_t))
            f.pop(index)
            c.pop(index)
        FF[i]=f
        CC[i]=c
    return FF,CC

def classify(measure,s_st,target=None,predictexe=True,n_cluster=2):
    
    num = measure.n//2+measure.n%2
    name = ''
    for i in measure.qubitToread:
        name += i
    if target is not None:
        name = f'q{target+1}'
    fig, axes = plt.subplots(ncols=2,nrows=num,figsize=(9,4*num))
    n = measure.n if target is None else 1
    if predictexe:
        for i in range(n):
            i = target if target is not None else i
            # s_off, s_on = s_st[0,:,i], s_st[1,:,i]
            # S = list(s_off) + list(s_on)
            # x,z = np.real(S), np.imag(S)
            # d = list(zip(x,z))
            # kmeans = KMeans(n_clusters=n_cluster,max_iter=1000,tol=0.0001)
            # kmeans.fit(d)
            x,z, predict,center,color=Classify_fit(s_st,qnum=i,nstate=n_cluster)

            measure.predict[measure.qubitToread[i]] = predict
            # y = predict(d)
            y=color
            print(list(y).count(1)/len(y))
            ax = axes[i//2][i%2] if num>1 else axes[i]
            ax.scatter(x,z,c=y,s=10)
            ax.axis('equal')
            ax.set_title(f'p0={list(y).count(0)/len(y)},p1={list(y).count(1)/len(y)}')
        plt.savefig(r'D:\skzhao\fig\%s.png'%(name+'predict'))
        plt.close()

        fig, axes = plt.subplots(ncols=2,nrows=num,figsize=(9,4*num))
        for i in range(n):
            i = target if target is not None else i
            s_off, s_on = s_st[0,:,i], s_st[1,:,i]
            ss, which = s_on, 0
            d = list(zip(np.real(ss),np.imag(ss)))
            y = measure.predict[measure.qubitToread[i]](d)
            percent1 = list(y).count(which)/len(y)
            # measure.which[measure.qubitToread[i]]={'g':0,'e':1,'f':2}
            measure.onwhich[measure.qubitToread[i]] = (which if percent1 > 0.5 else 1-which)
            measure.offwhich[measure.qubitToread[i]] = (1-which if percent1 > 0.5 else which)
            percent_on = list(y).count(measure.onwhich[measure.qubitToread[i]])/len(y)
            ax = axes[i//2][i%2] if num>1 else axes[i]
            ax.scatter(np.real(ss),np.imag(ss),c=y,s=10)
            ax.set_title(f'|1>pop={round(percent_on*100,3)}%')
            ax.axis('equal')
        plt.savefig(r'D:\skzhao\fig\%s.png'%(name+'e'))
        plt.close()

        fig, axes = plt.subplots(ncols=2,nrows=num,figsize=(9,4*num))
        for i in range(n):
            i = target if target is not None else i
            s_off, s_on = s_st[0,:,i], s_st[1,:,i]
            ss, which = s_off, measure.offwhich[measure.qubitToread[i]]
            d = list(zip(np.real(ss),np.imag(ss)))
            y = measure.predict[measure.qubitToread[i]](d)
            percent_off = list(y).count(which)/len(y)
            measure.readmatrix[measure.qubitToread[i]] = np.mat([[percent_off,1-percent_on],[1-percent_off,percent_on]])
            ax = axes[i//2][i%2] if num>1 else axes[i]
            ax.scatter(np.real(ss),np.imag(ss),c=y,s=10)
            ax.set_title(f'|0>pop={round(percent_off*100,3)}%')
            ax.axis('equal')
        plt.savefig(r'D:\skzhao\fig\%s.png'%(name+'g'))
        plt.close()
    else:
        fig, axes = plt.subplots(ncols=2,nrows=num,figsize=(9,4*num))
        for i in range(n):
            i = target if target is not None else i
            s_off, s_on = s_st[0,:,i], s_st[1,:,i]
            ss, which = s_on, measure.onwhich[measure.qubitToread[i]] 
            d = list(zip(np.real(ss),np.imag(ss)))
            y = measure.predict[measure.qubitToread[i]](d)
            percent_on = list(y).count(which)/len(y)
            ax = axes[i//2][i%2] if num>1 else axes[i]
            ax.scatter(np.real(ss),np.imag(ss),c=y,s=10)
            ax.set_title(f'|1>pop={round(percent_on*100,3)}%')
            ax.axis('equal')
        plt.savefig(r'D:\skzhao\fig\%s.png'%(name+'classify'))
        plt.close()

def cutCircle(info_c0,info_c1,data):
        # print(data.shape)
        # data0, data1 = data[0,:], data[1,:]
        c0, r0 = info_c0[0], info_c0[1]
        c1, r1 = info_c1[0], info_c1[1]
        s0 = (data.real-c0.real)**2+(data.imag-c0.imag)**2
        s1 = (data.real-c1.real)**2+(data.imag-c1.imag)**2
        # idx = ((np.abs(s0)<r0**2) +  (np.abs(s1)<r1**2))
        s1_cut = (s1<r1**2)*1
        s0_cut = (s0<r0**2)*1
        num0 = np.count_nonzero(s0_cut)
        num1 = np.count_nonzero(s1_cut)
        Na = num0+num1
        # print(num0)
        num0 /= Na
        num1 /= Na

        return num0, num1, Na

def P_depart(c0,c1,radio):
    x1,y1 = c0[0],c0[1]   
    x2,y2 = c1[0],c1[1]
    x = (x1 + radio*x2)/(1 + radio)
    y = (y1 + radio*y2)/(1 + radio)
    return x,y


def new_circle(c0,c1,r0,r1,rd=0.9):
    ###   rd < 1  ,,cross
    r01 = np.abs((c0[0]-c1[0]) + 1j*c0[1]-1j*c1[1])
    radio_0 = r0/(r01-r0)
    radio_1 = r1/(r01-r1)
    p_0 = P_depart(c0,c1,radio_0)
    p_1 = P_depart(c1,c0,radio_1)

    c_0 = P_depart(c0,p_0,-rd)
    c_1 = P_depart(c1,p_1,-rd)

    r_0 = np.abs((c_0[0]-p_0[0]) + 1j*c_0[1]-1j*p_0[1])
    r_1 = np.abs((c_1[0]-p_1[0]) + 1j*c_1[1]-1j*p_1[1])
    
    return c_0,c_1,r_0,r_1

def new_circle_2(c0,c1,r0,r1,rd=0.9):
    ###   rd < 1   ,,  no_cross
    r01 = np.abs((c0[0]-c1[0]) + 1j*c0[1]-1j*c1[1])
    radio_0 = r0/(r01-r0)
    radio_1 = r1/(r01-r1)
    p_0 = P_depart(c0,c1,radio_0)
    p_1 = P_depart(c1,c0,radio_1)

    c_0 = P_depart(c1,p_0,-rd)
    c_1 = P_depart(c0,p_1,-rd)

    r_0 = np.abs((c_0[0]-p_0[0]) + 1j*c_0[1]-1j*p_0[1])
    r_1 = np.abs((c_1[0]-p_1[0]) + 1j*c_1[1]-1j*p_1[1])
    
    return c_1,c_0,r_1,r_0

def circle_selcet(c0,c1,r0,r1,option='no_cross',rd=0.9):
    c0 = [c0.real,c0.imag]
    c1 = [c1.real,c1.imag]

    r01 = np.abs((c0[0]-c1[0]) + 1j*c0[1]-1j*c1[1])
    
    if option =='no_cross':

        if r01 < r0+r1:
            c0,c1,r0,r1 = new_circle_2(c0,c1,r0,r1,rd=rd)
        if r01 > r0+r1:
            c0,c1,r0,r1 = new_circle(c0,c1,r0,r1,rd=rd)
    if option =='cross':
        if r01 < r0+r1:
            c0,c1,r0,r1 = new_circle(c0,c1,r0,r1,rd=rd)
        if r01 > r0+r1:
            c0,c1,r0,r1 = new_circle_2(c0,c1,r0,r1,rd=rd)

    if option == 'tangency':
        RR = np.abs(c0 - c1)/(r0+r1)
        r0 = r0*RR
        r1 = r1*RR
        
    c0 = c0[0] + 1j*c0[1]
    c1 = c1[0] + 1j*c1[1]

    return c0,c1,r0,r1

# def circle_selcet(center0,center1,offstd, onstd, option='no_cross'):
    
#     if option == 'tangency':
#         RR = np.abs(center0 - center1)/(offstd+onstd)
#         offstd = offstd*RR
#         onstd = onstd*RR
#     if option == 'no_cross':
#         center0,center1,offstd, onstd = new_circle(center0,center1,offstd, onstd)

#     if option == 'cross':
#         center0,center1,offstd, onstd = new_circle_2(center0,center1,offstd, onstd)

#     return center0,center1,offstd, onstd

def find_circle(s_st,target=0,stdnum=[[1,1]],n_clusters=2,figsize=(9,14),tangency='no_cross',qubits_read=[],rd=0.9):
    from collections.abc import Iterable
    dim = np.shape(s_st)[-1]
    target = target if isinstance(target,Iterable) else [target]
    fig, axes = plt.subplots(ncols=2,nrows=dim,figsize=figsize)
    cLst_m, num0Lst, num1Lst = [], [], []
    S0_N ,S1_N= [], []
    cLst, pLst = [], []
    count = 0
    for i,qi in enumerate(range(dim)):
        ax0 = axes[i][0] if dim>1 else axes[0]
        ax1 = axes[i][1] if dim>1 else axes[1]
        if qi not in target:
            s_off, s_on = s_st[0,:,qi], s_st[1,:,qi]
            center1, center0 = np.mean(s_off), np.mean(s_on)
            sdiff = complex((center1-center0))
            s0 = s_off / (sdiff/np.abs(sdiff))
            s1 = s_on / (sdiff/np.abs(sdiff))
            s0 = np.real(s0)
            s1 = np.real(s1)
            bins = 120

            bin0 = ax1.hist(s0,bins=bins)
            bin1 = ax1.hist(s1,bins=bins)
         
            ax0.plot(np.real(s_off),np.imag(s_off),'.')
            ax0.plot(np.real(s_on),np.imag(s_on),'.')
            ax0.axis('equal')
            plt.tight_layout()
        else:
            
            x,z, predict,center,color,p=Classify_fit(s_st,qnum=i,nstate=n_clusters)
            pLst.append(p)
            center = np.array([center[i][0]+1j*center[i][1] for i in range(n_clusters)])
            center0, center1 = center[:2]
            if n_clusters < 3:
                s_off, s_on = s_st[0,:,qi], s_st[1,:,qi]
                sdiff = complex((center1-center0))
                s0 = s_off / (sdiff/np.abs(sdiff))
                s1 = s_on / (sdiff/np.abs(sdiff))
                s0 = np.real(s0)
                s1 = np.real(s1)
                bins = 120
                mu = complex(center1)/ (sdiff/np.abs(sdiff))
                x0 = np.linspace(min(s0),max(s0),bins)
                x1_old = np.linspace(min(s1),max(s1),bins)
                lower, high = np.real(mu)-2*np.std(s1), np.real(mu)+2*np.std(s1)
                x1 = x1_old[x1_old>lower]

                bin0 = ax1.hist(s0,bins=bins)
                bin1 = ax1.hist(s1,bins=bins)
                # para0, func0 = op.Gaussian_Fit().fitGaussian(x0,bin0[0],fine=False)
                # ax1.plot(x0,func0(x0,para0.x))
                b1 = bin1[0][x1_old>lower]
                b1 = b1[x1<high]
                x1 = x1[x1<high]
                # para1, func1 = op.Gaussian_Fit().fitGaussian(x1,b1,fine=False)
                # ax1.plot(x1,func1(x1,para1.x))
            #     axes[0].set_title(f'center={center0,center1}')

                # offstd, onstd = stdnum[count][0]*np.abs(para0.x[2]), stdnum[count][1]*np.abs(para1.x[2])
                offstd, onstd = stdnum[count][0]*np.std(s_off), stdnum[count][1]*np.std(s_on)

                # RR = np.abs(center0 - center1)/(offstd+onstd)
                # if tangency:
                #     offstd = offstd*RR
                #     onstd = onstd*RR

                center0,center1,offstd, onstd = circle_selcet(center0,center1,offstd, onstd, option=tangency,rd=rd)

                theta = np.arange(0, 2*np.pi, 0.01)
                roff = center0.real + offstd * np.cos(theta)
                ioff = center0.imag + offstd * np.sin(theta)
                ron = center1.real + onstd * np.cos(theta)
                ion = center1.imag + onstd * np.sin(theta)
                ax0.plot(np.real(s_off),np.imag(s_off),'.')
                # ax0.plot(np.real(s_on2),np.imag(s_on2),'.',alpha=0.4)
                ax0.plot(np.real(s_on),np.imag(s_on),'.')
                ax0.plot(center1.real,center1.imag,'bo')
                ax0.plot(center0.real,center0.imag,'ro')
                ax0.plot(roff,ioff)
                ax0.plot(ron,ion)
                ax0.axis('equal')



                cLst_m = [(center0,offstd),(center1,onstd)]

                # cLst.append([c0Lst,c1Lst])
                num0, n01 ,Na0 = cutCircle((center0,offstd),(center1,onstd),s_off)
                n10, num1, Na1 = cutCircle((center0,offstd),(center1,onstd),s_on)
                S0_N.append(Na0)
                S1_N.append(Na1)
                num0Lst.append(num0)
                num1Lst.append(num1)
                plt.tight_layout()
            else:

                cLst_m = [(center[i],stdnum[count][i]*np.std(s_st[i,:,qi])) for i in range(n_clusters)]
                num0Lst, num1Lst = p[:2]
                for n_state in range(n_clusters):
                    s = s_st[n_state,:,qi]
                    ax0.plot(np.real(s),np.imag(s),'.')
                color = ['ko','wo','bo']
                for n_state in range(n_clusters):
                    s = s_st[n_state,:,qi]
                    theta = np.arange(0, 2*np.pi, 0.01)
                    roff = np.real(center[n_state]) + stdnum[count][n_state]*np.std(s) * np.cos(theta)
                    ioff = np.imag(center[n_state]) + stdnum[count][n_state]*np.std(s) * np.sin(theta)
                    ax0.plot(np.real(center[n_state]),np.imag(center[n_state]),color[n_state],label=str(n_state))
                    ax0.plot(roff,ioff)
                    ax0.axis('equal')
                ax0.legend()
            cLst.append(cLst_m)
            count += 1
    return cLst, num0Lst, num1Lst, fig, axes, pLst
    

def post_selection(measure,s,sigma_num=2):
    center0,center1 = [], []
    radius0, radius1 = [], []
    for i in measure.postSle:
        center0.append(np.complex(*(measure.postSle[i][0][0])))
        radius0.append(measure.postSle[i][0][1])
        center1.append(np.complex(*(measure.postSle[i][1][0])))
        radius1.append(measure.postSle[i][1][1])
    pop = []
    for i in range(np.shape(s)[1]):
        x0 = (np.abs(s[:,i,:]-np.array(center0))<np.array(radius0)*sigma_num)*1.0
        x1 = (np.abs(s[:,i,:]-np.array(center1))<np.array(radius1)*sigma_num)*1.0
#         y = x0 + x1
#         c = np.sum(y ,axis=1)
#         x0, x1 = x0[c==10,:], x1[c==10,:]
#         print(x0)
        pop_sum1 = np.sum(x1,axis=1)
#         pop_sum0 = np.sum(x0,axis=1)
        pop_q = x1[pop_sum1==5,:]
#         print(pop_q[:,8])
#         print(np.shape(pop_q))
        pop.append((np.count_nonzero(pop_q,axis=0)/len(pop_q)))
    return np.array(pop)

################################################################################
# 对比度
################################################################################

def visibility(n,s0,s1, center=None):
    theta = np.arange(0, 2*np.pi, 0.01)
    data = []
    for i in range(n):
        c0, c1 = (np.mean(s0), np.mean(s1)) if center is None else center
        s0 = s0 / ((c1-c0)/np.abs(c1-c0))
        s1 = s1 / ((c1-c0)/np.abs(c1-c0))
        s0 = np.real(s0)
        s1 = np.real(s1)
        bins = np.linspace(np.min(np.r_[s0,s1]), np.max(np.r_[s0,s1]), 61)
        y0,_ = np.histogram(s0, bins=bins)
        y1,_ = np.histogram(s1, bins=bins)
        inte0 = np.cumsum(y0)/np.sum(y0)
        inte1 = np.cumsum(y1)/np.sum(y0)
        inte_diff = np.cumsum(y0)/np.sum(y0) - np.cumsum(y1)/np.sum(y1)
        offstd, onstd = np.std(s0), np.std(s1)
        roff = np.real(c0) + offstd * np.cos(theta)
        ioff = np.imag(c0) + offstd * np.sin(theta)
        ron = np.real(c1) + onstd * np.cos(theta)
        ion = np.imag(c1) + onstd * np.sin(theta)
        data.append([inte0,inte1,inte_diff,(roff,ioff),(ron,ion)])
    return data

# def visibility(n,s0,s1, center=None):
#     from scipy.optimize import least_squares as ls, curve_fit, basinhopping as bh
#     def Two_Gaussian(x, amp0,amp1, mu0,mu1, sigma0,sigma1):
#         amp = np.array([amp0,amp1])
#         mu = np.array([mu0,mu1])
#         sigma = np.array([sigma0,sigma1])
#         return np.sum(amp*np.exp(-(x[:,None]-mu)**2/2/sigma**2),axis=-1)
    
#     theta = np.arange(0, 2*np.pi, 0.01)
#     data = []
#     for i in range(n):
#         c0, c1 = (np.mean(s0), np.mean(s1)) if center is None else np.array(center)[:,0]+1j*np.array(center)[:,1]
#         s0 = s0 / ((c1-c0)/np.abs(c1-c0))
#         s1 = s1 / ((c1-c0)/np.abs(c1-c0))
#         s0 = np.real(s0)
#         s1 = np.real(s1)
#         bins = np.linspace(np.min([s0,s1]), np.max([s0,s1]), 61)
#         y0,_ = np.histogram(s0, bins=bins)
#         y1,_ = np.histogram(s1, bins=bins)
#         inte0 = np.cumsum(y0)/np.sum(y0)
#         inte1 = np.cumsum(y1)/np.sum(y0)
#         inte_diff = np.cumsum(y0)/np.sum(y0) - np.cumsum(y1)/np.sum(y1)
#         offstd, onstd = np.std(s0), np.std(s1)
#         roff = np.real(c0) + offstd * np.cos(theta)
#         ioff = np.imag(c0) + offstd * np.sin(theta)
#         ron = np.real(c1) + onstd * np.cos(theta)
#         ion = np.imag(c1) + onstd * np.sin(theta)
#         data.append([inte0,inte1,inte_diff,(roff,ioff),(ron,ion)])
#         vmax = np.max(inte_diff)
#         thr, phi = bins[inte_diff==vmax], phi = np.angle(c1-c0)
        
#         center_new = (c0/((c1-c0)/np.abs(c1-c0)),c1/((c1-c0)/np.abs(c1-c0)))
#         popt0 ,pcov0 = curve_fit(Two_Gaussian,bins[:-1],y0,p0=[0,np.max(y1),np.array(center_new).real[0],np.array(center_new).real[1],np.std(y1),np.std(y1)])
#         popt1 ,pcov1 = curve_fit(Two_Gaussian,bins[:-1],y1,p0=[0,np.max(y1),np.array(center_new).real[0],np.array(center_new).real[1],np.std(y1),np.std(y1)])
#     return data, thr, phi, bins, y0, y1, popt0, pcov0, popt1, pcov1

import math

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import EngFormatter, LogFormatterSciNotation
from scipy.optimize import least_squares as ls, curve_fit, basinhopping as bh
def Two_Gaussian(x, amp0,amp1, mu0,mu1, sigma0,sigma1):
    amp = np.array([amp0,amp1])
    mu = np.array([mu0,mu1])
    sigma = np.array([sigma0,sigma1])
    return np.sum(amp*np.exp(-(x[:,None]-mu)**2/2/sigma**2),axis=-1)


def plotLine(c0, c1, ax, **kwargs):
    t = np.linspace(0, 1, 11)
    c = (c1 - c0) * t + c0
    ax.plot(c.real, c.imag, **kwargs)


def plotCircle(c0, r, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    s = c0 + r * np.exp(1j * t)
    ax.plot(s.real, s.imag, **kwargs)


def plotEllipse(c0, a, b, phi, ax, **kwargs):
    t = np.linspace(0, 1, 1001) * 2 * np.pi
    c = np.exp(1j * t)
    s = c0 + (c.real * a + 1j * c.imag * b) * np.exp(1j * phi)
    ax.plot(s.real, s.imag, **kwargs)


def plotDistribution(s0,
                     s1,
                     fig=None,
                     axes=None,
                     info=None,
                     hotThresh=10000,
                     logy=False):
    from waveforms.math.fit import get_threshold_info, mult_gaussian_pdf

    if info is None:
        info = get_threshold_info(s0, s1)
    else:
        info = get_threshold_info(s0, s1, info['threshold'], info['phi'])
    thr, phi = info['threshold'], info['phi']
    # visibility, p0, p1 = info['visibility']
    # print(
    #     f"thr={thr:.6f}, phi={phi:.6f}, visibility={visibility:.3f}, {p0}, {1-p1}"
    # )

    if axes is not None:
        ax1, ax2 = axes
    else:
        if fig is None:
            fig = plt.figure()
        ax1 = fig.add_subplot(121)
        ax2 = fig.add_subplot(122)

    if (len(s0) + len(s1)) < hotThresh:
        ax1.plot(np.real(s0), np.imag(s0), '.', alpha=0.8)
        ax1.plot(np.real(s1), np.imag(s1), '.', alpha=0.8)
    else:
        _, *bins = np.histogram2d(np.real(np.hstack([s0, s1])),
                                  np.imag(np.hstack([s0, s1])),
                                  bins=50)

        H0, *_ = np.histogram2d(np.real(s0),
                                np.imag(s0),
                                bins=bins,
                                density=True)
        H1, *_ = np.histogram2d(np.real(s1),
                                np.imag(s1),
                                bins=bins,
                                density=True)
        vlim = max(np.max(np.abs(H0)), np.max(np.abs(H1)))

        ax1.imshow(H1.T - H0.T,
                   alpha=(np.fmax(H0.T, H1.T) / vlim).clip(0, 1),
                   interpolation='nearest',
                   origin='lower',
                   cmap='coolwarm',
                   vmin=-vlim,
                   vmax=vlim,
                   extent=(bins[0][0], bins[0][-1], bins[1][0], bins[1][-1]))

    ax1.axis('equal')
    # ax1.set_xticks([])
    # ax1.set_yticks([])
    for s in ax1.spines.values():
        s.set_visible(False)

    # c0, c1 = info['center']
    # a0, b0, a1, b1 = info['std']
    params = info['params']
    r0, i0, r1, i1 = params[0][0], params[1][0], params[0][1], params[1][1]
    a0, b0, a1, b1 = params[0][2], params[1][2], params[0][3], params[1][3]
    c0 = (r0 + 1j * i0) * np.exp(1j * phi)
    c1 = (r1 + 1j * i1) * np.exp(1j * phi)
    phi0 = phi + params[0][6]
    phi1 = phi + params[1][6]
    plotEllipse(c0, 2 * a0, 2 * b0, phi0, ax1)
    plotEllipse(c1, 2 * a1, 2 * b1, phi1, ax1)

    im0, im1 = info['idle']
    lim = min(im0.min(), im1.min()), max(im0.max(), im1.max())
    t = (np.linspace(lim[0], lim[1], 3) + 1j * thr) * np.exp(-1j * phi)
    ax1.plot(t.imag, t.real, 'k--')

    ax1.plot(np.real(c0), np.imag(c0), 'o', color='C3')
    ax1.plot(np.real(c1), np.imag(c1), 'o', color='C4')

    re0, re1 = info['signal']
    x, a, b, c = info['cdf']

    xrange = (min(re0.min(), re1.min()), max(re0.max(), re1.max()))

    n0, bins0, *_ = ax2.hist(re0, bins=100, range=xrange, alpha=0.5)
    n1, bins1, *_ = ax2.hist(re1, bins=100, range=xrange, alpha=0.5)

    x_range = np.linspace(x.min(), x.max(), 1001)
    *_, cov0, cov1 = info['std']

    c0, c1 = info['center']
    c0 /= ((c1-c0)/np.abs(c1-c0))
    c1 /= ((c1-c0)/np.abs(c1-c0))
    # print(len(n0),len(bins0))
    try:
        popt0 ,pcov0 = curve_fit(Two_Gaussian,bins0[:-1],n0,p0=[np.max(n0),0,c0.real,c1.real,np.std(n0),np.std(n1)])
        popt1 ,pcov1 = curve_fit(Two_Gaussian,bins1[:-1],n1,p0=[0,np.max(n1),c0.real,c1.real,np.std(n0),np.std(n1)])
        y0new = Two_Gaussian(x_range,*popt0)
        y1new = Two_Gaussian(x_range,*popt1)
    except:
        popt0 ,pcov0 = 0,0
        popt1 ,pcov1 = 0,0
        y0new = np.zeros_like(x_range)
        y1new = np.zeros_like(x_range)

    ax2.plot(
        x_range,
        y0new)
    ax2.plot(
        x_range,
        y1new)
    ax2.set_ylabel('Count')
    ax2.set_xlabel('Projection Axes')
    if logy:
        ax2.set_yscale('log')
        ax2.set_ylim(0.1, max(np.sum(n0), np.sum(n1)))

    ax3 = ax2.twinx()
    ax3.plot(x, a, '--', lw=1, color='C0')
    ax3.plot(x, b, '--', lw=1, color='C1')
    ax3.plot(x, c, 'k--', alpha=0.5, lw=1)
    ax3.set_ylim(0, 1.1)
    ax3.vlines(thr, 0, 1.1, 'k', alpha=0.5)
    ax3.set_ylabel('Probility')

    return info, [[popt0 ,pcov0],[popt1 ,pcov1]]




import numpy as np

def coherence_limit(nQ=2, T1_list=None, T2_list=None, gatelen=0.1):
        """
        This function is from qiskit
        The error per gate (1-average_gate_fidelity) given by the T1,T2 limit.

        Args:
            nQ (int): Number of qubits (1 and 2 supported).
            T1_list (list): List of T1's (Q1,...,Qn).
            T2_list (list): List of T2's (as measured, not Tphi). If not given assume T2=2*T1 .
            gatelen (float): Length of the gate.

        Returns:
            float: coherence limited error per gate.
        Raises:
            ValueError: If there are invalid inputs
        """
        # pylint: disable = invalid-name

        T1 = np.array(T1_list)

        if T2_list is None:
            T2 = 2 * T1
        else:
            T2 = np.array(T2_list)

        if len(T1) != nQ or len(T2) != nQ:
            raise ValueError("T1 and/or T2 not the right length")

        coherence_limit_err = 0

        if nQ == 1:

            coherence_limit_err = 0.5 * (
                1.0 - 2.0 / 3.0 * np.exp(-gatelen / T2[0]) - 1.0 / 3.0 * np.exp(-gatelen / T1[0])
            )

        elif nQ == 2:

            T1factor = 0
            T2factor = 0

            for i in range(2):
                T1factor += 1.0 / 15.0 * np.exp(-gatelen / T1[i])
                T2factor += (
                    2.0
                    / 15.0
                    * (
                        np.exp(-gatelen / T2[i])
                        + np.exp(-gatelen * (1.0 / T2[i] + 1.0 / T1[1 - i]))
                    )
                )

            T1factor += 1.0 / 15.0 * np.exp(-gatelen * np.sum(1 / T1))
            T2factor += 4.0 / 15.0 * np.exp(-gatelen * np.sum(1 / T2))

            coherence_limit_err = 0.75 * (1.0 - T1factor - T2factor)

        else:
            raise ValueError("Not a valid number of qubits")

        return coherence_limit_err


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   optimize.py
@Time    :   2020/02/10 16:55:28
@Author  :   sk zhao 
@Version :   1.0
@Contact :   2396776980@qq.com
@License :   (C)Copyright 2017-2018, Zhaogroup-iop
@Desc    :   None
'''

# here put the import lib
import time, numpy as np, matplotlib.pyplot as plt
from scipy import fftpack
from sklearn.cluster import KMeans
from scipy.optimize import least_squares as ls, curve_fit, basinhopping as bh
from scipy import signal
import asyncio, scipy
import sympy as sy
from collections import Counter
from collections.abc import Iterable
xvar = sy.Symbol('x',real=True)

'''
主要采用了全局优化的思想进行拟合
相比于optimize_old模块，该模块的思想是利用全局最优化算法进行数据拟合，方法采自scipy.optimize模块，以basinhopping为例，就是来最小化残差函数，
相比于最小二乘法，他要求残差函数返回的为每一点的残差平方和，该法对初始值相对不敏感，精度高。
'''


# func_willBeused = {'Expfunc':Exp_Fit().func}


################################################################################
### 拟合参数边界
################################################################################

class MyBounds(object):
    def __init__(self, xmax=[1.1,1.1], xmin=[-1.1,-1.1] ):
        self.xmax = np.array(xmax)
        self.xmin = np.array(xmin)
    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmax = bool(np.all(x <= self.xmax))
        tmin = bool(np.all(x >= self.xmin))
        return tmax and tmin
        
################################################################################
### 生数据预处理  exeFit
###################### ##########################################################

class RowToRipe():
    def __init__(self):
        pass
    def space(self,y):
        ymean, ymax, ymin = np.mean(y), np.max(y), np.min(y)
        d = y[np.abs(y-ymean)>0.05*(ymax-ymin)]
        if len(d) < 0.9*len(y):
            return len(d) + int(len(y)*0.1)
        else:
            return len(y)
    
    def errspace(self,func,paras,args):

        return (func(args['x'],paras)-args['y'])**2

    def deductPhase(self,x,y):
        if np.ndim(y) != 2:
            y = [y]
        s = []
        for i in y:
            phi = np.unwrap(np.angle(i), 0.9 * np.pi)
            phase = np.poly1d(np.polyfit(x, phi, 1))
            base = i / np.exp(1j * phase(x))
            s.append(base)
        return x, np.array(s)

    def manipulation(self,volt,freq,s,which='min',dosmooth=False,f0=0.25,axis=1):
        manfunc = np.argmin if which == 'min' else np.argmax
        s_abs = np.array(s) 
         
        if dosmooth: 
            s_abs = self.smooth(s_abs,axis=axis,f0=f0)
            # min_index = []      
            # for i in range(np.shape(s)[axis]):
            #     z = s_abs[i,:]
            #     # loc = freq[manfunc(z)]
            #     z_smooth = self.movingAverage(z,smoothnum)
            #     # t,index = self.firstMax(freq,z_smooth,num=loc,peakpercent=0.9,insitu=True)
            #     index = manfunc(z_smooth)
            #     min_index.append(index)
        # else:
        #     min_index = manfunc(s_abs,axis=axis)     
        min_index = manfunc(s_abs,axis=axis)         
        x, y = np.array(volt), np.array([freq[j] for j in min_index]) 
        return x,y

    def extension(self,s,ax=0,periodic_extension=0,padding=0):
        axis = np.shape(s)

        if len(axis) == 2:
            if periodic_extension:
                if ax == 1:
                    s = np.array((list(s)+list(s[:,::-1]))*periodic_extension)
                if ax == 0:
                    s = np.array((list(np.array(s).T)+list(np.array(s[::-1,:]).T))*periodic_extension).T
            shape = (axis[0]*2*periodic_extension+padding,axis[1]) if ax == 1 else (axis[0],axis[1]*2*periodic_extension+padding)
            data = np.zeros(shape)
            rows, cols = np.shape(s)
            print(shape)
            data[:rows,:cols] = s
        else:
            if periodic_extension:
                if ax == 0:
                    s = np.array((list(s)+list(s[::-1]))*periodic_extension)
            shape = (axis[0]*2*periodic_extension+padding,)
            data = np.zeros(shape)
            rows = len(s)
            data[:rows] = s
        return data
  
    def firstMax(self,x,y,num=0,peakpercent=0.9,insitu=False,mean=True,which='max'):
        """
        x:       线性扫描点
        insitu:  如果为True，截取peak值为x为num时对应的值乘以peakpercent，False为最大值乘以peakpercent
        """
        if which != 'max':
            y = np.abs(y-y.max())
        index0 = np.argmin(np.abs(x-num))
        y = y - np.min(y)
        peak = peakpercent*y[index0] if insitu else peakpercent*np.max(y)
        c = np.argwhere(y>peak)
        cdiff = np.diff(c[:,0])
        n_clusters = len(np.argwhere(cdiff>np.mean(cdiff))) + 1
        S = c[:,0]
        d = np.asarray(np.mat(list(zip(S,S))))

        kmeans = KMeans(n_clusters=n_clusters,max_iter=100,tol=0.001)
        yfit = kmeans.fit_predict(d)
        xaxis = S[yfit==yfit[np.argmin(np.abs(S-index0))]]
        index =  int(np.mean(xaxis)) if mean else int(xaxis[np.argmax(y[xaxis])])
        bias0 = round(x[index],5)
        return bias0 ,index

    def smooth(self,y,f0=0.1,axis=-1):
        b, a = signal.butter(3,f0)
        z = signal.filtfilt(b,a,y,axis=axis)
        return z

    def movingAverage(self,data,num2mean=5,ratio=1):
        x = np.exp(np.arange(num2mean))
        window = (x / np.sum(x))
    #     window = np.ones(num2mean)/num2mean
        ynew1 = np.convolve(data, window, 'same')
        ynew1[:num2mean] = data[:num2mean]
        ynew2 = np.convolve(data[::-1], window, 'same')
        ynew2[:num2mean] = data[::-1][:num2mean]
        ynew = (ynew1+ynew2[::-1])/2
        return ynew

    def resample(self,x,y,num=1001):
        down = len(x)
        up = num
        x_new = np.linspace(min(x),max(x),up)
        z = signal.resample_poly(y,up,down)
        return x_new, z

    def findPeaks(self,y,width=None,f0=0.015,h=0.15,threshold=None,prominence=None,plateau_size=None,rel_height=0):
        detrend = np.mean(y - signal.detrend(y))
        # mask = y > (np.max(y)+np.min(y))/2
        z = y if np.max(y)-detrend>detrend-np.min(y) else -y
        background = self.smooth(z,f0=f0)
        height0 = (np.max(z)-np.min(z))
        height = (background+h*height0,background+(1+h)*height0)
        threshold = threshold if threshold == None else threshold*height0
        property_peaks = signal.find_peaks(z,height=height,threshold=threshold,plateau_size=plateau_size)
        index = property_peaks[0]
        half_widths = signal.peak_widths(z,index,rel_height=rel_height)
        print(index,half_widths[0])
        # side = (index+int(half_widths[0]), index-int(half_widths[0]))
        side = 0
        prominence = signal.peak_prominences(z,index)
        return index, side, prominence
    
    def spectrum(self,x,y,method='normal',window='boxcar',detrend='constant',axis=-1,scaling='density',average='mean',shift=True):
        '''
        scaling:
            'density':power spectral density V**2/Hz
            'spcetrum': power spectrum V**2
        '''
        fs = (len(x)-1)/(np.max(x)-np.min(x))
        if method == 'normal':
            f, Pxx = signal.periodogram(y,fs,window=window,detrend=detrend,axis=axis,scaling=scaling)
        if method == 'welch':
            f, Pxx = signal.welch(y,fs,window=window,detrend=detrend,axis=axis,scaling=scaling,average=average)
        f, Pxx = (np.fft.fftshift(f), np.fft.fftshift(Pxx)) if shift else (f, Pxx)
        index = np.argmax(Pxx,axis=axis)
        w = f[index]
        return w, f, Pxx
    
    def cross_psd(self,x,y,z,window='hann',detrend='constant',scaling='density',axis=-1,average='mean'):
        fs = (len(x)-1)/(np.max(x)-np.min(x))
        f, Pxy = signal.csd(y,z,fs,window=window,detrend=detrend,scaling=scaling,axis=axis,average=average)
        return f, Pxy
    
    def ftspectrum(self,x,y,window='hann',detrend='constant',scaling='density',axis=-1,mode='psd'):
        '''
        mode:
            'psd':
            'complex':==stft
            'magnitude':==abs(stft)
            'angle':with unwrapping
            'phase':without unwraping
        '''
        fs = (len(x)-1)/(np.max(x)-np.min(x))
        f, t, Sxx = signal.spectrigram(y,fs,window=window,detrend=detrend,scaling=scaling,axis=axis,mode=mode)
        return f, t, Sxx
    
    def stft(self,x,y,window='hann',detrend=False,axis=-1,boundary='zeros',padded=True,nperseg=256):
        '''
        boundary:you can choose ['even','odd','constant','zeros',None]
        padded: True Or False          
        '''
        fs = (len(x)-1)/(np.max(x)-np.min(x))
        f, t, Zxx = signal.stft(y,fs,window=window,detrend=detrend,axis=axis,boundary=boundary,padded=padded,nperseg=nperseg)
        return f, t, Zxx
     
    def istft(self,x,Zxx,window='hann',boundary=True,time_axis=-1,freq_axis=-2):
        fs = (len(x)-1)/(np.max(x)-np.min(x))
        t, y = signal.stft(Zxx,fs,window=window,boundary=boundary,time_axis=time_axis,freq_axis=freq_axis)
        return t, y


    def fourier(self,x,y,axis=-1,detrend=True,zeroPadding=0,cycleExtension=1,window='boxcar',printinfo=False):
        '''

        x: linear time array or sample_rate
        y: 1d or 2d array 
        axis: FFT along the axis
        detrend: detrend the offset of y if True ,default True.
        zeroPadding: append zero for y along the axis
        cycleExtension: cycleExtend y along the axis
        window: str or tuple. Desired window to use. 
                it is passed to get_window to generate the window values, 
                which are DFT-even by default. See get_window for a list of 
                windows and required parameters. Defaults to a boxcar window.
        '''
        
        y = signal.detrend(y,axis=axis,type='constant') if detrend else y
        shape = list(np.shape(y))
        shape[axis] = zeroPadding
        
        window = signal.get_window(window,np.shape(y)[axis])  
        window = np.expand_dims(window,axis=axis-1).repeat(shape[axis-1],axis=axis-1) if len(shape)!=1 else window
        
        sample = (np.max(x) - np.min(x))/(len(x) - 1) if isinstance(x,Iterable) else x
        if printinfo:
            print(f'Computation resolution:{sample/np.shape(y)[axis]}')

        y0 = np.concatenate([y*window]*cycleExtension,axis=axis)

        ynew = np.concatenate([y0,np.zeros(shape)],axis=axis)
        if printinfo:
            print(f'Computation resolution after zeroPadding:{sample/(np.shape(ynew)[axis]-np.shape(y0)[axis]+np.shape(y)[axis])}')
            print(f'Computation resolution after resample:{sample/np.shape(ynew)[axis]}')

        yt  = np.fft.rfftfreq(np.shape(ynew)[axis],sample)
        amp = np.fft.rfft(ynew,axis=axis)
        w = np.abs(yt[np.argmax(np.abs(amp),axis=axis)])

        return w, yt, amp
        
    def envelope(self,y,responsetime=100):
        mold, out, rc = 0, [], responsetime
        out.append(np.abs(y[0]))
        for j, i in enumerate(y[1:],start=1):
            i = np.abs(i)
            if i > out[j-1]:
                mold = i
            else:
                mold = (out[j-1] * rc)/(rc + 1)
            out.append(mold)
        return out

    def envelope_Hilbert(self,y,axis=0):
        ym = signal.detrend(y,type='constant',axis=axis)
        yh = signal.hilbert(ym,axis=axis) 
        out = np.abs(ym + 1j*yh) + y.mean()
        return out

    def freq_Hilbert(self,x,y,axis=0):
        ym = signal.detrend(y,type='constant',axis=axis)
        yh = signal.hilbert(ym,axis=axis) 
        phase = np.unwrap(np.angle(ym+1j*yh))
        res, func = self.poly(x,phase,1)
        w = res[0]
        return w, phase, func

    def profile(self,v,f,s,peak,axis=1,classify=False):
        if classify:
            index = np.argwhere(np.abs(s)>peak)
            v = v[index[:,0]]
            f = f[index[:,1]]
        else:
            if axis == 1:
                v = v[np.abs(s).max(axis=1)>peak]
                s = s[np.abs(s).max(axis=1)>peak]
                f = f[np.abs(s).argmax(axis=1)]
            if axis == 0:
                f = f[np.abs(s).max(axis=0)>peak]
                s = s[:,np.abs(s).max(axis=0)>peak]
                v = v[np.abs(s).argmax(axis=0)]
        return v, f

    def profile1(self,v,f,s,peak,axis=1,classify=False):
        
        v = v[np.abs(s).max(axis=1)>peak]
        s = s[np.abs(s).max(axis=1)>peak]
        f = f[np.abs(s).argmax(axis=1)]
        if axis == 0:
            f = f[np.abs(s).max(axis=0)>peak]
            s = s[:,np.abs(s).max(axis=0)>peak]
            v = v[np.abs(s).argmax(axis=0)]
        return v, f

    def poly(self,x,y,num=1):
        z = np.polyfit(x, y, num)
        func = np.poly1d(z)
        return z, func

    def interp1d(self,x,y,kind='cubic',axis=-1, copy=True, bounds_error=None, fill_value=np.nan, assume_sorted=False):
        '''
        interp1d(x, y, kind='linear', axis=-1, copy=True, bounds_error=None, fill_value=nan, assume_sorted=False)
        |  
        |  Interpolate a 1-D function.
        |  
        |  `x` and `y` are arrays of values used to approximate some function f:
        |  ``y = f(x)``. This class returns a function whose call method uses
        |  interpolation to find the value of new points.
        |  
        |  Parameters
        |  ----------
        |  x : (N,) array_like
        |      A 1-D array of real values.
        |  y : (...,N,...) array_like
        |      A N-D array of real values. The length of `y` along the interpolation
        |      axis must be equal to the length of `x`.
        |  kind : str or int, optional
        |      Specifies the kind of interpolation as a string or as an integer
        |      specifying the order of the spline interpolator to use.
        |      The string has to be one of 'linear', 'nearest', 'nearest-up', 'zero',
        |      'slinear', 'quadratic', 'cubic', 'previous', or 'next'. 'zero',
        |      'slinear', 'quadratic' and 'cubic' refer to a spline interpolation of
        |      zeroth, first, second or third order; 'previous' and 'next' simply
        |      return the previous or next value of the point; 'nearest-up' and
        |      'nearest' differ when interpolating half-integers (e.g. 0.5, 1.5)
        |      in that 'nearest-up' rounds up and 'nearest' rounds down. Default
        |      is 'linear'.
        |  axis : int, optional
        |      Specifies the axis of `y` along which to interpolate.
        |      Interpolation defaults to the last axis of `y`.
        |  copy : bool, optional
        |      If True, the class makes internal copies of x and y.
        |      If False, references to `x` and `y` are used. The default is to copy.
        |  bounds_error : bool, optional
        |      If True, a ValueError is raised any time interpolation is attempted on
        |      a value outside of the range of x (where extrapolation is
        |      necessary). If False, out of bounds values are assigned `fill_value`.
        |      By default, an error is raised unless ``fill_value="extrapolate"``.
        |  fill_value : array-like or (array-like, array_like) or "extrapolate", optional
        |      - if a ndarray (or float), this value will be used to fill in for
        |        requested points outside of the data range. If not provided, then
        |        the default is NaN. The array-like must broadcast properly to the
        |        dimensions of the non-interpolation axes.
        |      - If a two-element tuple, then the first element is used as a
        |        fill value for ``x_new < x[0]`` and the second element is used for
        |        ``x_new > x[-1]``. Anything that is not a 2-element tuple (e.g.,
        |        list or ndarray, regardless of shape) is taken to be a single
        |        array-like argument meant to be used for both bounds as
        |        ``below, above = fill_value, fill_value``. Using a two-element tuple
        |        or ndarray requires ``bounds_error=False``.
        |  
        |        .. versionadded:: 0.17.0
        |      - If "extrapolate", then points outside the data range will be
        |        extrapolated.
        |  
        |        .. versionadded:: 0.17.0
        |  assume_sorted : bool, optional
        |      If False, values of `x` can be in any order and they are sorted first.
        |      If True, `x` has to be an array of monotonically increasing values.

        return 0, func_interp

        '''
        func = scipy.interpolate.interp1d(x,y,kind=kind,axis=axis, copy=copy, bounds_error=bounds_error, fill_value=fill_value, assume_sorted=assume_sorted)

        return 0, func

################################################################################
### 拟合Exp函数
################################################################################

class Exp_Fit(RowToRipe):
    
    def __init__(self,funcname=None):
        self.funcname = funcname

    def func(self,x,paras):

        if self.funcname == 'gauss':
            A, B, T1, T2 = paras
            return A * np.exp(-T2*x**2-x*T1) + B 
        else:
            A, B, T1 = paras
            return A * np.exp(-x*T1) + B 
    
    def errExp(self,paras, x, y):
        
        if self.funcname == 'gauss':
            return np.sum((self.func(x,paras) - y)**2)
        else:
            return np.sum((self.func(x,paras) - y)**2)

    def guessExp(self,x,y):
        ymin = y.min()
        y = y-y.min()
        mask = y > 0.05*y.max()
        if self.funcname == 'gauss':
            a = np.polyfit(x[mask], np.log(y[mask]), 2)
            return [np.exp(a[2]), ymin, -a[1], -a[0]]
        else:
            a = np.polyfit(x[mask], np.log(y[mask]), 1)
            return [np.exp(a[1]), ymin, -a[0]]

    def fitExp(self,x,y):
        p0 = self.guessExp(x,y)
        # res = ls(self.errExp, p0, args=(x, y)) 
        res = bh(self.errExp,p0,niter = 500,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)}) 
        return res, self.func
func_willBeused = {'Expfunc':Exp_Fit().func}
################################################################################
### 拟合Gaussian函数
################################################################################

class Gaussian_Fit(RowToRipe):
    
    def __init__(self):
        pass

    def func(self,x,paras):
        A, B, sigma, mu = paras
        return A/sigma/np.sqrt(2*np.pi)*np.exp(-(x-mu)**2/2/sigma**2) + B

    def errGaussian(self,paras, x, y,weight):
        return np.sum(weight*(self.func(x,paras) - y)**2)

    def errGaussian_ls(self,paras, x, y,weight):
        return weight*(self.func(x,paras) - y)

    def guessGaussian(self,x,y):
        background = np.mean(y - signal.detrend(y,type='constant'))
        height = (np.max(y)+np.min(y))/2
        B = y.min()
        mu = np.mean(x[y>height]) if y[len(y)//2] > background else np.mean(x[y<height])
        sigma = (np.max(x[y>height]) - np.min(x[y>height])) if y[len(y)//2] > background else (np.max(x[y<height]) - np.min(x[y<height])) 
        A = (np.max(y)-np.min(y))*sigma*np.sqrt(2*np.pi) if y[len(y)//2] > background else -(np.max(y)-np.min(y))*sigma*np.sqrt(2*np.pi)
        return [A, B, sigma,mu]

    def fitGaussian(self,x_old,y_old,correct=False,fine=True):
        l = len(y_old)
        reslst, cost = [], []
        idxLst, rLst = ([0,l//6,-l//6],[1,-1]) if fine else ([0],[1])
        for i in idxLst:
            x , y = (x_old[i:], y_old[i:]) if i >=0 else (x_old[:i], y_old[:i])
            for ratio in rLst:
                A, B, sigma, mu = self.guessGaussian(x,y)
                # weight = (self.func(x,(A,0,sigma,mu))/np.max(self.func(x,(A,0,sigma,mu))))**10
                weight = 1
                A *= ratio
                p0 = A, B, sigma, mu
                # print(p0)
                # mybounds = MyBounds(xmin=[-10,-10,-np.inf,-np.inf,-np.inf,-np.inf],xmax=[10,10,-np.inf,-np.inf,-np.inf,-np.inf])
                res = bh(self.errGaussian,p0,niter = 200,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y,weight)})    
                reslst.append(res)
                cost.append(res.fun)
        index = np.argmin(cost)
        if correct:
            res = ls(self.errGaussian_ls,reslst[index].x,args=(x_old, y_old,weight))
            return res, self.func
        else:
            return reslst[index], self.func

################################################################################
### 拟合双Gaussian函数
################################################################################

class Gaussian2_Fit(RowToRipe):
    
    def __init__(self):
        pass

    def func(self,x,paras):
        amp, mu, sigma = paras
        return np.sum(amp*np.exp(-(x[:,None]-mu)/2/sigma**2),axis=-1)

    def errGaussian(self,paras, x, y,weight):
        return np.sum(weight*(self.func(x,paras) - y)**2)

    def errGaussian_ls(self,paras, x, y,weight):
        return weight*(self.func(x,paras) - y)

    def guessGaussian(self,x,y):
        background = np.mean(y - signal.detrend(y,type='constant'))
        height = (np.max(y)+np.min(y))/2
        B = y.min()
        mu = np.mean(x[y>height]) if y[len(y)//2] > background else np.mean(x[y<height])
        sigma = (np.max(x[y>height]) - np.min(x[y>height])) if y[len(y)//2] > background else (np.max(x[y<height]) - np.min(x[y<height])) 
        A = (np.max(y)-np.min(y))*sigma*np.sqrt(2*np.pi) if y[len(y)//2] > background else -(np.max(y)-np.min(y))*sigma*np.sqrt(2*np.pi)
        return [A, B, sigma,mu]

    def fitGaussian(self,x_old,y_old,correct=False):
        l = len(y_old)
        reslst, cost = [], []
        for i in [0,l//6,-l//6]:
            x , y = (x_old[i:], y_old[i:]) if i >=0 else (x_old[:i], y_old[:i])
            for ratio in [1,-1]:
                A, B, sigma, mu = self.guessGaussian(x,y)
                # weight = (self.func(x,(A,0,sigma,mu))/np.max(self.func(x,(A,0,sigma,mu))))**10
                weight = 1
                A *= ratio
                p0 = A, B, sigma, mu
                # print(p0)
                # mybounds = MyBounds(xmin=[-10,-10,-np.inf,-np.inf,-np.inf,-np.inf],xmax=[10,10,-np.inf,-np.inf,-np.inf,-np.inf])
                res = bh(self.errGaussian,p0,niter = 200,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y,weight)})    
                reslst.append(res)
                cost.append(res.fun)
        index = np.argmin(cost)
        if correct:
            res = ls(self.errGaussian_ls,reslst[index].x,args=(x_old, y_old,weight))
            return res, self.func
        else:
            return reslst[index], self.func

################################################################################
### 拟合Cos函数
################################################################################

class Cos_Fit(RowToRipe):

    def __init__(self,phi=None):
        self.phi = phi
        pass

    def func(self,x,paras):
        A,C,W,phi = paras  
        return A*np.cos(2*np.pi*W*x+phi)+C

    def errCos(self,paras,x,y,kind):      
        if kind == 'bh':       
            return  np.sum((self.func(x,paras)-y)**2)  
        if kind == 'ls':
            return self.func(x,paras) - y

    def guessCos(self,x,y):
        x, y = np.array(x), np.array(y)
        # sample = (np.max(x) - np.min(x))/(len(x) - 1)
        Ag, Cg= np.abs(y-np.mean(y)).max(), np.mean(y) 
        # yt  = np.fft.fftshift(np.fft.fftfreq(len(y))) / sample
        # amp = np.fft.fftshift(np.fft.fft(y))
        Wg,yt,amp = RowToRipe().fourier(x, y)
        z = np.abs(amp[yt!=0])
        ytz = yt[yt!=0]
        # Wg = np.abs(ytz[np.argmax(z)])
        phig =  np.mean(np.arccos((y - Cg)/Ag) - 2*np.pi*Wg*x) % (2*np.pi)
        return Ag, Cg, Wg, 0

    def fitCos(self,volt,s):
        x, y = volt, s
        if x[0] / 1e9 > 1:
            raise 'I hate the large number, please divided by 1e9, processing x in GHz'
        Ag, Cg, Wg, phig = self.guessCos(x,y)
        phig = phig if self.phi is None else self.phi
        p0 = Ag, Cg, Wg, phig
        # print(Ag, Cg, Wg, phig)
        # res = ls(self.errCos, [Ag,Cg,Wg,phig], args=(x, y)) 
        mybounds = MyBounds(xmin=[-np.inf,-np.inf,0,-np.pi],xmax=[np.inf,np.inf,1.5*Wg,np.pi])    
        res = bh(self.errCos,p0,niter=80,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y,'bh')},accept_test=mybounds)   
        res = ls(self.errCos, res.x, args=(x, y,'ls')) 

        return res, self.func

################################################################################
### 拟合Cosh函数
################################################################################

class Cosh_Fit(RowToRipe):

    def __init__(self,A=None,C=None,w=None,phi=None):
        self.phi = phi
        self.A = A
        self.C = C
        self.w = w
        pass

    def func(self,x,paras):
        A,C,W,phi = paras  
        return A*np.cosh(W*x+phi)+C

    def errCosh(self,paras,x,y,kind):      
        if kind == 'bh':       
            return  np.sum((self.func(x,paras)-y)**2)  
        if kind == 'ls':
            return self.func(x,paras) - y

    def guessCosh(self,x,y):
        x, y = np.array(x), np.array(y)
        C = y - signal.detrend(y,type='constant')
        y = signal.detrend(y,type='constant')
    
        mask = y > 0.05*y.max()
        a = np.polyfit(x[mask], np.log(y[mask]), 1)
        A, w = np.exp(a[1]), -a[0]
        
        w = w if self.w is None else self.w
        phi = phi if self.phi is None else self.phi
        A = A if self.A is None else self.A
        C = C if self.C is None else self.C
        return A, C, w, phi

    def fitCosh(self,volt,s):
        x, y = volt, s
        # if x[0] / 1e9 > 1:
        #     raise 'I hate the large number, please divided by 1e9, processing x in GHz'
        Ag, Cg, Wg, phig = self.guessCosh(x,y)
        p0 = Ag, Cg, Wg, phig
        # print(Ag, Cg, Wg, phig)
        # res = ls(self.errCos, [Ag,Cg,Wg,phig], args=(x, y)) 
        mybounds = MyBounds(xmin=[-np.inf,-np.inf,0,-np.inf],xmax=[np.inf,np.inf,10*Wg,np.inf])    
        res = bh(self.errCosh,p0,niter=80,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y,'bh')},accept_test=mybounds)   
        res = ls(self.errCosh, res.x, args=(x, y,'ls')) 

        return res, self.func

################################################################################
### 拟合洛伦兹函数
################################################################################

class Lorentz_Fit(RowToRipe):
    '''
    I hate the large number
    processing x in GHz
    '''
    def __init__(self):
        pass

    def func(self,x,paras):
        a,b,c,d = paras
        return a/(1.0+c*(x-b)**2)+d

    def errLorentz(self,paras,x,y):
        return np.sum((self.func(x,paras)-y)**2)

    def guessLorentz(self,x,y):
        # index, prominences, widths = self.findPeaks(y)
        z = np.sort(np.abs(y))
        d = np.mean(z[:int(len(z)/2)])
        # y = np.abs(y)- d
        b = x[np.abs(y).argmax()]
        # b1, b = x[index]
        bw = (np.max(x[y>0.5*(np.max(y)+np.min(y))])-np.min(x[y>0.5*(np.max(y)+np.min(y))]))/2
        print(bw/1e6)
        # bw1, bw = widths
        a = np.abs(y-d).max()
        # a1, a = prominences
        c = 1 / bw**2
        return a,b,c,d

    def fitLorentz(self,x,y):
        if x[0] / 1e9 > 1:
            raise 'I hate the large number, please divided by 1e9, processing x in GHz'
        para = self.guessLorentz(x,y)
        print(para)
        # mybounds = MyBounds(xmin=[-np.inf,-np.inf,-np.inf,-np.inf,0,0],xmax=[np.inf,np.inf,np.inf,np.inf,1.5*w,2*np.pi])    
        res = bh(self.errLorentz,para,niter=20,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)})
        # res = ls(self.errLorentz,para,args=(x,y))
        a,b,c,d = res.x
        return res,self.func,np.sqrt(np.abs(1/c))*2e3

################################################################################
### 拟合T1
################################################################################

class T1_Fit():
    "A, T1"
    def __init__(self,T1_limt=100e3,b_lim=0.):
        self.T1_limt = T1_limt
        self.b_lim = b_lim
    
    def func(self,x,paras):
        A, T1,B = paras
        f = A * np.exp(-x/T1) +B*self.b_lim
        return f

    def errT1(self,paras, x, y):
        return np.sum((self.func(x,paras) - y)**2)

    def errT1_ls(self,paras, x, y):
        return self.func(x,paras) - y

    def errT1_cf(self,x,A, T1,B):
        paras = A, T1,B
        return self.func(x,paras) 

    def guessT1(self,x,y):
        A = y.max()
        y = y/A
        y = RowToRipe().smooth(y,f0=0.25)
        B = y.min()
        y = y-y.min()
        # mask = y > y.min()
        a = np.polyfit(x, np.log(y), 1)
        T1 = np.abs(1/a[0]) if np.abs(1/a[0]) < self.T1_limt else 0.99*self.T1_limt
        # A = np.exp(a[1]) if np.exp(a[1]) > 0.95 and np.exp(a[1])<1.1 else 1
        # A, T1, B = 1,x[np.abs(y-np.exp(-1))<0.1][0], 0
        return [A, T1,B]

    def fitT1(self,x,y,s=None):
        """[fit T1 or 2d-T1]

        Args:
            x ([1d-array]): [if s is None,x is time, else voltage or frequency of qubit in 2d-T1]
            y ([1d-array]): [if s is None,y is population of |1>, else time in 2d-T1]
            s ([2d-array]): [population of |1>]. Defaults to None.

        Returns:
            [res, func]: [if s is None]
            [res_lst]: [if s is not None]
        """
        if s is None:
            p0 = self.guessT1(x,y)
            print(p0)
            mybounds = MyBounds(xmin=[0.95,0,0],xmax=[1.1,self.T1_limt,self.b_lim])    
            res = bh(self.errT1,p0,niter = 50,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds) 
            return res,self.func,0
            # res = ls(self.errT1_ls,p0,args=(x,y))
            # popt,pcov = curve_fit(self.errT1_cf,x,y,p0,bounds=([0.95,0,0],[1.1,self.T1_limt,0.25]))
            # return popt, self.func, np.sqrt(np.diag(pcov))
        else:
            T1_lst = []
            pcov_lst = []
            for i in range(len(x)):
                data = s[i,:]/np.max(s[i,:])
                p0 = self.guessT1(y,data)
                mybounds = MyBounds(xmin=[0.95,0,0],xmax=[1.1,self.T1_limt,self.b_lim])    
                res = bh(self.errT1,p0,niter = 50,minimizer_kwargs={"method":"Nelder-Mead","args":(y,s[i,:])},accept_test=mybounds) 
                T1_lst.append(res.x[1])
                pcov_lst.append(res.fun)
                # res = ls(self.errT1_ls,p0,args=(y,s[i,:]))
                # popt,pcov = curve_fit(self.errT1_cf,y,s[i,:],p0,bounds=([0.95,0,0],[1.1,self.T1_limt,0.25]))
                # T1_lst.append(popt[1])
                # pcov_lst.append(np.sqrt(np.diag(pcov)))
            return T1_lst, pcov_lst

################################################################################
### 拟合指数包络函数
################################################################################

class T2_Fit(Exp_Fit,Cos_Fit):
    '''
    #############
    example:
    import imp
    import optimize
    op = imp.reload(optimize)
    try: 
        fT2 = op.T2_Fit(funcname='gauss',envelopemethod='hilbert')
        A,B,T1,T2,w,phi = fT2.fitT2(t,y)
    finally:
        pass
    ##############
    '''
    def __init__(self,responsetime=100,T1=35000,phi=0,funcname=None,envelopemethod=None):
        Exp_Fit.__init__(self,funcname)
        self.responsetime = responsetime
        self.T1 = T1
        self.phi = phi
        self.envelopemethod = envelopemethod
    
    def guessT2(self,x,y_new,y):
 
        res, _ = self.fitExp(x[5:-5],y_new[5:-5])
        A, B, T1, T2 = res.x
        T1 = 1 / T1 / 2
        if self.T1 is not None:
            T1 = self.T1
        Ag, Cg, Wg, phig = self.guessCos(x,y)
        return A, B, T1, np.sqrt(np.abs(1/T2)), Wg, phig

    def func_T2(self,x,para):
        A,B,T1,T2,w,phi = para
        return A*np.exp(-(x/T2)**2-x/T1/2)*np.cos(2*np.pi*w*x+phi) + B

    def errT2(self,para,x,y):
        return np.sum((self.func_T2(x,para) - y)**2)

    def fitT2(self,x,y):
        '''
        几个参数的限制范围还需要考究，A，T1，T2
        '''
        d = self.space(y)
        if self.envelopemethod == 'hilbert':
            out = self.envelope_Hilbert(y)
        else:
            out = self.envelope(y)
        A,B,T1,T2,w,phi = self.guessT2(x,out,y)
        env = A,B,T1,T2,out
        if T2 > 0.8*x[d-1] and d < 0.8*len(y):
            T2 = 0.37*x[d-1]
        amp = (np.max(y)-np.min(y)) / 2
        A = A if np.abs(A-amp) < 0.1*amp else amp
        p0 = A,B,T1,T2,w,self.phi
        print(p0)
        # res = ls(self.errT2, p0, args=(x, y)) 
        mybounds = MyBounds(xmin=[0,-np.inf,0,0,0,-np.pi],xmax=[np.inf,np.inf,100000,100000,1.5*w,np.pi])    
        res = bh(self.errT2,p0,niter = 30,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds)     
        A,B,T1,T2,w,phi = res.x
        return res, self.func_T2

class T2Envelope_Fit(Exp_Fit,Cos_Fit):
    '''
    #############
    example:
    import imp
    import optimize
    op = imp.reload(optimize)
    try: 
        fT2 = op.T2_Fit(funcname='gauss',envelopemethod='hilbert')
        A,B,T1,T2,w,phi = fT2.fitT2(t,y)
    finally:
        pass
    ##############
    '''
    def __init__(self,funcname=None):
        Exp_Fit.__init__(self,funcname)
    
    def guessT2Envelope(self,x,y):
 
        res, _ = self.fitExp(x,y)
        A, B, T1, T2 = res.x
        T1 = 1 / T1 / 2
        return A, B, T1, np.sqrt(np.abs(1/T2))

    def func_T2Envelope(self,x,para):
        A,B,T1,T2 = para
        return A*np.exp(-(x/T2)**2-x/T1/2) + B

    def errT2Envelope(self,para,x,y):
        return np.sum((self.func_T2Envelope(x,para) - y)**2)

    def fitT2Envelope(self,x,y):
        '''
        几个参数的限制范围还需要考究，A，T1，T2
        '''
        A,B,T1,T2 = self.guessT2Envelope(x,y)
        # env = A,B,T1,T2,out
        # if T2 > 0.8*x[d-1] and d < 0.8*len(y):
        #     T2 = 0.37*x[d-1]
        # amp = (np.max(y)-np.min(y)) / 2
        # A = A if np.abs(A-amp) < 0.1*amp else amp
        p0 = A,B,T1,T2
        print(p0)
        # res = ls(self.errT2, p0, args=(x, y)) 
        mybounds = MyBounds(xmin=[0,-np.inf,0,0],xmax=[np.inf,np.inf,100000,100000])    
        res = bh(self.errT2Envelope,p0,niter = 80,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds)     
        A,B,T1,T2 = res.x
        return res, self.func_T2Envelope

class Rabi_Fit(T2_Fit):

    def __init__(self,responsetime=100,T1=20000,phi=np.pi/2,funcname=None,envelopemethod=None):
        T2_Fit.__init__(self,responsetime,T1,phi,funcname,envelopemethod)
        
    
    def guessRabi(self,x,y_new,y):
 
        res, _ = self.fitExp(x[5:-5],y_new[5:-5])
        A, B, T1 = res.x
        T1 = 1 / T1
        if self.T1 is not None:
            T1 = self.T1
        Ag, Cg, Wg, phig = self.guessCos(x,y)
        return A, B, T1, Wg, phig
    
    def func_Rabi(self,x,paras):
        A,B,T1,w,phi = paras
        return A*np.exp(-x/T1)*np.cos(2*np.pi*w*x+phi) + B

    def errRabi(self,paras,x,y):
        
        return np.sum((self.func_Rabi(x,paras) - y)**2)

    def fitRabi(self,x,y):
        if self.envelopemethod == 'hilbert':
            out = self.envelope_Hilbert(y)
        else:
            out = self.envelope(y)
        A,B,T1,w,phi = self.guessRabi(x,out,y)
        env = (A,B,T1,out)
        amp = (np.max(y)-np.min(y)) / 2
        A = A if np.abs(A-amp) < 0.1*amp else amp
        B = B if np.abs(B-np.mean(y)) < 0.1*np.mean(y) else np.mean(y)
        p0 = A,B,T1,w,self.phi
        print(p0)
        # res = ls(self.errRabi, p0, args=(np.array(x), np.array(y)))   
        mybounds = MyBounds(xmin=[0,-np.inf,100,0,0],xmax=[np.inf,np.inf,100e3,1.5*w,2*np.pi])
        res = bh(self.errRabi,p0,niter=30,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds)      
        A,B,T1,w,phi = res.x
        return res, self.func_Rabi

################################################################################
### 拟合二维谱
################################################################################


class Spec2d_Fit(Cos_Fit):

    def __init__(self,peak=15,threshold=0.001,voffset=None, vperiod=None, ejs=None, ec=0.2, d=1):
        Cos_Fit.__init__(self,phi=None)
        self.peak = peak
        self.threshold = threshold
        self.voffset = voffset
        self.vperiod = vperiod
        self.ejs = ejs
        self.ec = ec
        self.d = d
    
    def profile(self,v,f,s,classify=False):
        if classify:
            index = np.argwhere(s>self.peak)
            v = v[index[:,0]]
            f = f[index[:,1]]
        else:
            v = v[s.max(axis=1)>self.peak]
            # s = s[s.max(axis=1)>self.peak]
            # f = f[s.argmax(axis=1)]
            idx = s > self.peak
            f = [f[i][-1] for i in idx if np.any(i) == True]
        return v, np.array(f)
    def func_f01(self,x,paras):
        voffset, vperiod, ejs, ec, d = paras
        tmp = np.pi*(x-voffset)/vperiod
        f01 = np.sqrt(8*ejs*ec) * np.abs(np.cos(tmp)) * np.sqrt(1+d**2*np.tan(tmp)**2) - ec
        return f01
    def err(self,paras,x,y):
        # A, C, w, phi = paras
        # f01 = np.sqrt(A*np.abs(np.cos(w*x+phi))) + C
        f01 = self.func_f01(x,paras)
        return np.sum((f01 - y)**2)

    def fitSpec2d(self,v,f,s=None,classify=False):
        if s is not None:
            v,f = self.profile(v,f,s,classify)
            # print(list(v),list(f))
        W, _, _ = self.fourier(v,f,zeroPadding=10000)
        try:
            voffset = self.firstMax(v,f,num=0)[0]
        except:
            voffset  =  v[np.argmax(f)]
            
        vperiod = np.abs(2/W)
        d, ec = self.d, self.ec
        ejs = np.abs((np.max(f)+ec)**2/8/ec)
        
        voffset = voffset if self.voffset is None else self.voffset
        vperiod = vperiod if self.vperiod is None else self.vperiod
        ejs = ejs if self.ejs is None else self.ejs
        p0 = [voffset, vperiod,ejs,ec,d]
        # res, func = self.fitCos(v,f)
        # space = np.abs(func(v,res.x)-f)
        # if np.max(space) > self.threshold:
        #     # print(space)
        #     v = v[space<self.threshold]
        #     f = f[space<self.threshold]

        while 1:
            print(p0)
            # print(list(v),list(f))
            xmin = [voffset-0.5*np.abs(voffset),0,0,0,0]
            xmax = [voffset+0.5*np.abs(voffset),1.5*vperiod,2*ejs,2*ec,1000]
            mybounds = MyBounds(xmin=xmin,xmax=xmax)
            res = bh(self.err,p0,niter = 200,minimizer_kwargs={"method":"Nelder-Mead","args":(v, f)},accept_test=mybounds) 
            # res = ls(self.err,p0,args=(v, f),bounds=(xmin,xmax)) 
            # voffset, vperiod, ejs, ec, d = res.x
            space = self.errspace(self.func_f01,res.x,{'x':v,'y':f})
            if np.max(space) > self.threshold:
                # print(space)
                v = v[space<self.threshold]
                f = f[space<self.threshold]
                p0 = res.x
                # print(len(v),(space<0.001))
            else:
                return f, v, res, self.func_f01
################################################################################
### 拟合腔频调制曲线
################################################################################

class Cavitymodulation_Fit(Spec2d_Fit):

    def __init__(self,peak=15,phi=None):
        Cos_Fit.__init__(self,phi=phi)
        self.peak = peak

    def func_s(self,x,paras):
        voffset, vperiod, ejs, ec, d, g, fc = paras
        tmp = np.pi*(x-voffset)/vperiod
        f01 = np.sqrt(8*ejs*ec*np.abs(np.cos(tmp))*np.sqrt(1+d**2*np.tan(tmp)**2))-ec
        fr = (fc+f01+np.sqrt(4*g**2+(f01-fc)**2))/2
        # fr = fc - g**2/(f01-fc)
        return fr

    def err(self,paras,x,y):
        return np.sum((self.func_s(x,paras) - y)**2)

    def fitCavitymodulation(self,v,f,s,classify=False):
        
        '''
        v:voltage,
        f:frequency,
        s:data
        '''
        shape = np.shape(s)
        axis = np.where(np.array(shape)!=len(v))[0][0]
        v,f = self.manipulation(v,f,s,axis=axis)
        paras, func = self.fitCos(v,f)
        A, C, W, phi = paras.x
        voffset, vperiod, ec, d= self.firstMax(v,f,num=0)[0], 1/W, 0.1*np.min(f), 1
        # g = np.min(f)-fc
        ejs = (np.max(f)+ec)**2/8/ec
        g, fc = ec, np.mean(f)
        p0 = [voffset, vperiod, ejs, ec, d, g, fc]
        print(p0)
        mybounds = MyBounds(xmin=[-0.25*vperiod,0,0,0,0,0,0],xmax=[0.25*vperiod,1.5*vperiod,2*ejs,2*ec,2,2*g,2*fc])
        res = bh(self.err,p0,niter = 200,minimizer_kwargs={"method":"Nelder-Mead","args":(v, f)},accept_test=mybounds)
        # res = ls(self.err,res.x,args=(v, f)) 
        # A, C, W, phi = res.x
        voffset, vperiod, ejs, ec, d, g, fc = res.x
        return f, v, res, self.func_s

################################################################################
### crosstalk直线拟合
################################################################################

class Crosstalk_Fit(Spec2d_Fit):

    def __init__(self,peak=15):
        self.peak = peak

    def two_line(self,f,v):
        fv = []
        for i,j in enumerate(f):
            fv.append([j,v[i]])
        fv = sorted(fv)
        ff = sorted(list(Counter(f).keys()))
        F = [0]*len(ff)
        for i,j in enumerate(ff):
            F_m=[]
            for k,l in enumerate(fv):
                if j==l[0]:
                    F_m.append(l[1])
            F[i]=sorted(F_m )

            
        FF_1 = [0]*len(F)
        FF_2 = [0]*len(F)
        FF_m = [0]*len(F)
        for k,i in enumerate(F):
            F_1 = []
            F_2 = []
            for j in i:
                if j<i[-1]-0.05:
                    F_2.append(j)
                else:
                    F_1.append(j)
            if len(F_2)==len(i):
                FF_m[k]=F_2.copy()
                F_2 = []
            if len(F_1)==len(i):
                FF_m[k]=F_1.copy()
                F_1 = []
                
            FF_1[k] = F_1
            FF_2[k] = F_2

        vm=[]
        for j, i in enumerate(FF_m):
            if i==0:
                vm.append(j)

        FF_m1 = (FF_m[:vm[0]])
        v_m1 = (ff[:vm[0]])
        FF_m2 = (FF_m[vm[-1]+1:])
        v_m2 = (ff[vm[-1]+1:])
                
        c1=[]
        v1=[]
        for j, i in enumerate(FF_1):
            c1+=len(i)*[1*ff[j]]
            v1+=i
            
        c2=[]
        v2=[]
        for j, i in enumerate(FF_2):
            c2+=len(i)*[1*ff[j]]
            v2+=i


        cm1=[]
        vm1=[]
        for j, i in enumerate(FF_m1):
            cm1+=len(i)*[1*v_m1[j]]
            vm1+=i
            
        cm2=[]
        vm2=[]
        for j, i in enumerate(FF_m2):
            cm2+=len(i)*[1*v_m2[j]]
            vm2+=i


        p1 = np.array([np.polyfit(c1, v1 ,1)[1],np.polyfit(c2, v2 ,1)[1]])-np.polyfit(cm1, vm1 ,1)[1]
        p2 = np.array([np.polyfit(c1, v1 ,1)[1],np.polyfit(c2, v2 ,1)[1]])-np.polyfit(cm2, vm2 ,1)[1]

        if abs(p2[1])<abs(p2[0]):
            v2+=vm2
            c2+=cm2
        if abs(p2[1])>abs(p2[0]):
            v1+=vm2
            c1+=cm2

        if abs(p1[1])<abs(p1[0]):
            v2+=vm1
            c2+=cm1
        if abs(p1[1])>abs(p1[0]):
            v1+=vm1
            c1+=cm1
        return v1,c1,v2,c2
        
    def fitCrosstalk(self,v,f,s,classify=False):
        v,f = self.profile(v,f,s,classify)
        res = np.polyfit(f,v,1)
        return v, f, res

################################################################################
### 单比特tomo
################################################################################

def pTorho(plist):
    pz_list, px_list, py_list = plist
    rho_list = []
    for i in range(np.shape(pz_list)[0]):
        pop_z, pop_x, pop_y = pz_list.T[i], px_list.T[i], py_list.T[i]
        rho_00, rho_01 = 1 - pop_z, (2*pop_x - 2j*pop_y - 1 + 1j) / 2j
        rho_10, rho_11 = (1 + 1j - 2*pop_x - 2j*pop_y) / 2j, pop_z
        rho = np.array([[rho_00,rho_01],[rho_10,rho_11]])
        rho_list.append(rho)
    pass

################################################################################
### RB
################################################################################

class RB_Fit:
    def __init__(self,A=None,B=None):
        self.A = A
        self.B = B
        pass
    def func_rb(self,x,paras):
        A,B,p = paras
        B = B if self.B is None else self.B
        return A*p**x+B
    def err(self,paras,x,y):
        return self.func_rb(x,paras)-y
    def guess(self,x,y):
        B = np.min(y) if self.B is None else self.B
        y = y - np.min(y)
        mask = y > 0
        a = np.polyfit(x[mask], np.log(y[mask]), 1)
        A = np.exp(np.abs(a[1])) if self.A is None else self.A
        p = 1/np.exp(np.abs(a[0]))
        return A, B, p
    def fitRB(self,x,y):
        p0 = self.guess(x,y)

        res = ls(self.err, p0, args=(x, y)) 
        A,B,p = res.x
        return res, self.func_rb

################################################################################
### 双指数拟合
################################################################################

# class TwoExp_Fit(Exp_Fit):
#     def __init__(self,funcname=None,percent=0.2):
#         Exp_Fit.__init__(self,funcname)
#         self.percent = percent
#     def err(self,paras,x,y):
#         a, b, c, d, e = paras
#         return np.sum((a*np.exp(b*x) + c*np.exp(d*x) + e - y)**2)
#     def guess(self,x,y):
#         a,e,b = self.fitExp(x,y)
#         b *= -1
#         e = np.min(y) if a > 0 else np.max(y)
#         return a,b,a*self.percent,b*self.percent,e
#     def fitTwoexp(self,x,y):
#         p0 = self.guess(x,y)
#         a, b, c, d, e = p0
#         lower = [0.95*i if i > 0 else 1.05*i for i in p0]
#         higher = [1.05*i if i > 0 else 0.95*i for i in p0]
#         lower[2], lower[3] = -np.abs(a)*self.percent, -np.abs(b)*self.percent
#         higher[2], higher[3] = self.percent*np.abs(a), self.percent*np.abs(b)
#         print(p0)
#         # res = ls(self.err,p0,args=(x,y),bounds=(lower,higher))
#         res = bh(self.err,p0,niter = 50,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)})

#         return res.x

class TwoExp_Fit(Exp_Fit):
    def __init__(self,funcname=None,percent=0.2):
        Exp_Fit.__init__(self,funcname)
        self.percent = percent
    def fitfunc(self,x,p):
        return (p[0] + np.sum(p[1::2,None]*np.exp(-p[2::2,None]*x[None,:]), axis=0))
    def err(self,paras,x,y):
        return np.sum(((self.fitfunc(x,paras) - y)*(1.0+0.5*scipy.special.erf(0.4*(x-paras[2])))**5)**2)
    def guess(self,x,y,paras):
        offset = np.min(y)
        alist = np.max(y) - np.min(y)
        blist = np.max(x)-np.min(x)
        paras[0] = offset
        paras[1::2] = alist
        paras[2::2] = 1/blist
        return paras
    def fitTwoexp(self,x,y,num=2):
        paras = np.zeros((2*num+1,))
        xmin, xmax = np.zeros((2*num+1,)), np.zeros((2*num+1,))
        p0 = self.guess(x,y,paras)
        xmin[0], xmax[0] = p0[0]*0.5, p0[0]*1.5
        xmin[1::2], xmin[2::2] = p0[1::2]*0.5, -(np.max(x)-np.min(x))*2
        xmax[1::2], xmax[2::2] = p0[1::2]*1.5, (np.max(x)-np.min(x))*2
        mybounds = MyBounds(xmin=xmin,xmax=xmax)
        res = bh(self.err,p0,niter = 50,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)})
        # res = bh(self.err,res.x,niter = 50,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds)
        p0 = res.x
        print(xmin)
        # res = ls(self.err, p0, args=(np.array(x), np.array(y)))  
        return res.x

################################################################################
## 误差函数拟合
################################################################################

class Erf_fit(RowToRipe):
    def __init__(self):
        RowToRipe.__init__(self)
    def func(self,x,paras):
        sigma1, sigma2, center1, center2, a, b = paras
        return a*(scipy.special.erf((x-center1)/sigma1)+np.abs(scipy.special.erf((x-center2)/sigma2)-1))+b
    def err(self,paras,x,y):
        return np.sum((y-self.func(x,paras))**2)
    def guess(self,x,y):
        height = np.max(y) - np.min(y)
        mask = x[y < (np.max(y)+np.min(y))/2]
        center1, center2 = mask[-1], mask[0]
        b = np.mean(y - signal.detrend(y,type='constant'))
        a = (np.max(y) - np.min(y)) if y[len(y)//2] < np.mean(y) else -(np.max(y) - np.min(y))
        z, ynew = x[(np.min(y)+0.1*height)<y], y[(np.min(y)+0.1*height)<y]
        z = z[ynew<(np.max(ynew)-0.2*height)]
        sigma2 = (z[z<np.mean(z)][-1]-z[z<np.mean(z)][0])
        sigma1 = (z[z>np.mean(z)][-1]-z[z>np.mean(z)][0])
        return sigma1, sigma2, center1, center2, a, b
    def fitErf(self,x,y,printinit=True):
        l = len(y)
        reslst, cost = [], []
        for i in [0,l//6,-l//6]:
            x , y = (x[i:], y[i:]) if i >=0 else (x[:i], y[:i])
            for ratio in [1,-1]:
                sigma1, sigma2, center1, center2, a, b = self.guess(x,y)
                a *= ratio
                paras = sigma1, sigma2, center1, center2, a, b
                if printinit:
                    print(paras) 
                # mybounds = MyBounds(xmin=[-10,-10,-np.inf,-np.inf,-np.inf,-np.inf],xmax=[10,10,-np.inf,-np.inf,-np.inf,-np.inf])
                res = bh(self.err,paras,niter = 100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)}) 
                reslst.append(res)
                cost.append(res.fun)
        index = np.argmin(cost)
        return reslst[index], self.func

class Erf_fit1(RowToRipe):
    def __init__(self):
        RowToRipe.__init__(self)
    def func(self,x,paras):
        sigma1, sigma2, center1, center2, a1, b1, a2, b2 = paras
        return a1*scipy.special.erf((x-center1)/sigma1)+b1+a2*np.abs(scipy.special.erf((x-center2)/sigma2)-1)+b2
    def err(self,paras,x,y):
        return np.sum((y-self.func(x,paras))**2)
    def guess(self,x,y):
        height = np.max(y) - np.min(y)
        mask = x[y < (np.max(y)+np.min(y))/2]
        center1, center2 = mask[-1], mask[0]
        b = np.mean(y - signal.detrend(y,type='constant'))
        a = (np.max(y) - np.min(y)) if y[len(y)//2] < np.mean(y) else -(np.max(y) - np.min(y))
        z, ynew = x[(np.min(y)+0.1*height)<y], y[(np.min(y)+0.1*height)<y]
        z = z[ynew<(np.max(ynew)-0.2*height)]
        sigma2 = (z[z<np.mean(z)][-1]-z[z<np.mean(z)][0])
        sigma1 = (z[z>np.mean(z)][-1]-z[z>np.mean(z)][0])
        return sigma1, sigma2, center1, center2, a, b, a, b
    def fitErf(self,x,y,printinit=True):
        l = len(y)
        reslst, cost = [], []
        for i in [0,l//6,-l//6]:
            x , y = (x[i:], y[i:]) if i >=0 else (x[:i], y[:i])
            for ratio in [1,-1]:
                sigma1, sigma2, center1, center2, a1, b1, a2, b2 = self.guess(x,y)
                a1 *= ratio
                a2 *= ratio
                paras = sigma1, sigma2, center1, center2, a1, b1, a2, b2
                if printinit:
                    print(paras) 
                # mybounds = MyBounds(xmin=[-10,-10,-np.inf,-np.inf,-np.inf,-np.inf],xmax=[10,10,-np.inf,-np.inf,-np.inf,-np.inf])
                res = bh(self.err,paras,niter = 100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)}) 
                reslst.append(res)
                cost.append(res.fun)
        index = np.argmin(cost)
        return reslst[index], self.func

################################################################################
## 拟合单个误差函数
################################################################################

class singleErf_fit(RowToRipe):
    def __init__(self):
        RowToRipe.__init__(self)
    def func(self,x,paras):
        sigma1, center1, a, b = paras
        return a*scipy.special.erf((x-center1)/sigma1)+b
    def err(self,paras,x,y):
        return np.sum((y-self.func(x,paras))**2)
    def guess(self,x,y):
        mask = x[y < y.mean()]
        center1 = mask[-1]
        b = np.mean(y - signal.detrend(y))
        a = np.max(y) - np.min(y)
        z = np.abs(y - np.mean(y))
        xnew = x[z<(np.max(z)+np.min(z))/2]
        sigma1 = xnew[-1] - xnew[0]
        return sigma1, center1, a, b
    def fitErf(self,x,y):
        l = len(y)
        reslst, cost = [], []
        for i in [0,l//6,-l//6]:
            x , y = (x[i:], y[i:]) if i >=0 else (x[:i], y[:i])
            for ratio in [1,-1]:
                sigma1, center1, a, b = self.guess(x,y)
                a *= ratio
                paras = sigma1, center1, a, b
                # print(paras)
                res = bh(self.err,paras,niter = 100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)}) 
                reslst.append(res)
                cost.append(res.fun)
        index = np.argmin(cost)
        return reslst[index], self.func

################################################################################
## 拟合圆
################################################################################

class Circle_Fit():
    def __init__(self):
        pass
    def func(self,paras):
        xc, yc, R = paras
        theta = np.linspace(0,2*np.pi,1001)
        x = R*np.cos(theta) + xc
        y = R*np.sin(theta) + yc
        return x, y
    def errfunc(self,paras,x,y):
        xc, yc, R = paras
        return (x - xc)**2 + (y - yc)**2 - R**2
    def guess(self,x,y):
        xc = (np.max(x)+np.min(x))/2
        yc = (np.max(y)+np.min(y))/2
        R = np.sqrt((x-xc)**2+(y-yc)**2)
        return xc, yc, np.mean(R)
    def fitCircle(self,x,y):
        p0 = self.guess(x,y)
        res = ls(self.errfunc,p0,args=(x,y)) 
        return res, self.func

################################################################################
## 拟合贝塞尔函数绝对值
################################################################################

class Bessel_fit(RowToRipe):
    def __init__(self):
        RowToRipe.__init__(self)
    def func(self,x,paras):
        alpha, a = paras
        return a*np.abs(scipy.special.jv(0,alpha*x))
    def err(self,paras,x,y,kind='bh'):
        if kind == 'bh':
            return np.sum((y-self.func(x,paras))**2)
        if kind == 'ls':
            return y-self.func(x,paras)
    def guess(self,x,y):
        b = np.min(y)
        a = np.max(y)
        alpha = 2.4048/x[np.argmin(y)]
        return alpha, a
    def fitBessel(self,x,y):

        paras = self.guess(x,y)
        print(paras)
        # res = bh(self.err,paras,niter = 100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)}) 
        # return res, self.func
        while 1:
            # print(p0)
            res = bh(self.err,paras,niter = 100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y,'bh')}) 
            res = ls(self.err,res.x,args=(x,y,'ls')) 
            # space = self.errspace(self.func,res.x,{'x':x,'y':y})
            # if np.max(space) > 0.001:
            #     x = x[space<0.001]
            #     y = y[space<0.001]
            #     paras = res.x
                # print(len(v),(space<0.001))
            # else:
            return res, self.func

################################################################################
## 拟合zPulse核函数
################################################################################

# class zKernel_fit(RowToRipe):
#     def __init__(self,timeconstants =0,polyNum = 17,tCut = 90,height = 0.9,sigma = 0.4,tstart = 0,tshift = 0,polyRatio = 1,expNum = 1,tCut1 = 0):
#         self.polyNum = 17 #10  polyNum阶多项式拟合
#         self.expNum = 1 #######################目前固定为1,多项式拟合时的指数
#         self.timeconstants =0 # expNum阶指数拟合
#         self.tCut = 90#100 ,200  polyNUm的拟合范围
#         self.tCut1 = 0
#         self.tshift = 0
#         # self.height = eval(tags[1])
#         self.height = 0.9

#         self.polyRatio = 1
#         self.sigma = 0.4
#         self.tstart = 0

#     def fitfunc(t, p):
#         return (t >= 0) * ( np.sum(p[1::2,None]*(np.exp(-p[2::2,None]*(t[None,:]+tp))-np.exp(-p[2::2,None]*t[None,:]))/p[2::2,None], axis=0)) 
#     def errfunc(p,t,phi):
#         return (fitfunc(t, p)* height * diff_height -phi )*(1.0+0.5*scipy.special.erf(0.4*(t-buffer)))**10

#     def err(p,t,phi):
#         return np.sum(errfunc(p,t,phi)**2)

#     p0 = np.zeros(2*timeconstants+1)
#     p0[1::2] = -np.linspace(0, 0.02, timeconstants)
#     p0[2::2] = np.linspace(0, 0.5, timeconstants)
#     # print(p0)
#     p, _ = scipy.optimize.leastsq(errfunc,p0,maxfev=5000)
#     ts = np.arange(0,data[-1,0]*5,0.02)
#     p[0] = 0

#     plt.figure()
#     plt.subplot(211)
#     plt.plot(data[:,0],data[:,1],'bo')
#     plt.plot(ts,fitfunc(ts,p),'k-')

#     data1 = np.copy(data)
#     restData = data[:,1]-fitfunc(data[:,0], p)
#     restData = restData[data1[:,0]<=tCut]
#     data1 = data1[data1[:,0]<=tCut,:]
#     def fitfunc1(t, p):
#         pExp = p[:expNum*2]
#         pPoly = p[expNum*2:]
#         if np.iterable(t):
#             return np.sum(pExp[0::2,None]*np.exp(-pExp[1::2,None]*t[None,:]), axis=0)*np.polyval(pPoly,t)*(t<=tCut+20)
#         else:
#             return np.sum(pExp[0::2]*np.exp(-pExp[1::2]*t))*np.polyval(pPoly,t)*(t<=tCut)
#     def errfunc1(p):
#         return (fitfunc1(data1[:,0], p) - restData)
#     def smoothFuncATtCut(ts,tCut,tshift):    
#         return (0.5-0.5*scipy.special.erf(sigma*(ts-tCut+tshift)))*(0.5+0.5*scipy.special.erf(4.0*(ts-data[0,0]+0.5)))
#     pExp0 = np.zeros(2*expNum)
#     pExp0[1::2] = np.linspace(0.001, 0.03, expNum)
#     pPoly0 = np.zeros(polyNum)
#     pPoly0[-1] = 1.0
#     pAll0 = np.hstack([pExp0,pPoly0])
#     p2, _ = scipy.optimize.leastsq(errfunc1, pAll0)
#     smoothData = fitfunc1(ts,p2)*smoothFuncATtCut(ts,tCut,tshift)*polyRatio
#     timeFunData0 = smoothData+fitfunc(ts,p)
#     paras= {'p':p,'p2':p2,'tCut':tCut,'tshift':tshift,'sigma':sigma}
#     plt.subplot(212)
#     plt.plot(data1[:,0],restData,'bo')
#     # plt.plot(ts,smoothData,'r-')
#     plt.plot(ts,fitfunc1(ts,p2))
#     plt.xlabel('Time (ns)')
#     plt.grid(True)
#     plt.subplot(211)
#     plt.plot(ts,timeFunData0,'r-')
#     plt.grid(True)
#     p2 = np.asarray(p2)
#     p = np.asarray(p)
#     p_uniform = np.copy(p)
#     p2_uniform = np.copy(p2)
#     p_uniform[1::2] = p_uniform[1::2]/height
#     pExp = p2_uniform[:expNum*2]
#     pPoly = p2_uniform[expNum*2:]
#     pExp[0::2] = pExp[0::2]/height*polyRatio
#     p2_uniform[:expNum*2] = pExp

################################################################################
## 真空拉比拟合
################################################################################

class Vcrabi_fit():
    def __init__(self,length=None):
        self.length = length
    def func(self,x,paras):
        g, A0, Z0 = paras
        return np.sqrt(4*(g)**2+A0**2*(x-Z0)**2)
    def err(self,paras,x,y):
        return np.sum((self.func(x,paras)-y)**2)
    def guess(self,x,y):
        Z0 = x[np.argmin(y)]
        g = np.min(y)/2
        x, y = x[x!=Z0], y[x!=Z0]
        A0 = np.mean(np.sqrt(y**2-4*(g)**2)/(x-Z0))
        return g, A0, Z0
    def fitVcrabi(self,x,y):
        if self.length is None:
            p0 = self.guess(x,y)
            mybounds = MyBounds(xmin=[0,0,-1.1],xmax=[0.1,np.inf,1.1])
            res = bh(self.err,p0,niter=100,minimizer_kwargs={"method":"Nelder-Mead","args":(x, y)},accept_test=mybounds)
            return res, self.func
        else:
            index = len(y)//2
            start, end = index-self.length, index+self.length
            x0, y0 = x[start:end], y[start:end]
            p0 = self.guess(x0,y0)
            mybounds = MyBounds(xmin=[0,0,-1.1],xmax=[0.1,np.inf,1.1])
            res = bh(self.err,p0,niter=100,minimizer_kwargs={"method":"Nelder-Mead","args":(x0, y0)},accept_test=mybounds)
            
            space = np.abs(y-self.func(x,res.x)) < 0.1*np.abs(self.func(x,res.x))
            x0, y0 = x[space], y[space]
            p0 = self.guess(x0,y0)
            mybounds = MyBounds(xmin=[0,0,-1.1],xmax=[0.1,np.inf,1.1])
            res = bh(self.err,p0,niter=100,minimizer_kwargs={"method":"Nelder-Mead","args":(x0, y0)},accept_test=mybounds)
            return res, self.func

################################################################################
## 拟合Q值
################################################################################

class Cavity_fit(RowToRipe):
    def __init__(self):
        pass

    def circleLeastFit(self,x, y):
        def circle_err(params, x, y):
            xc, yc, R = params
            return (x - xc)**2 + (y - yc)**2 - R**2

        p0 = [
            x.mean(),
            y.mean(),
            np.sqrt(((x - x.mean())**2 + (y - y.mean())**2).mean())
        ]
        res = ls(circle_err, p0, args=(x, y))
        return res.x

    def guessParams(self,x,s):
        
        y = np.abs(1 / s)
        f0 = x[y.argmax()]
        _bw = x[y > 0.5 * (y.max() + y.min())]
        FWHM = np.max(_bw) - np.min(_bw)
        Qi = f0 / FWHM
        _, _, R = self.circleLeastFit(np.real(1 / s), np.imag(1 / s))
        Qe = Qi / (2 * R)
        QL = 1 / (1 / Qi + 1 / Qe)

        return [f0, Qi, Qe, 0, QL]

    def invS21(self, f, paras):
        f0, Qi, Qe, phi = paras
        #QL = 1/(1/Qi+1/Qe)
        return 1 + (Qi / Qe * np.exp(1j * phi)) / (
            1 + 2j * Qi * (np.abs(f) / np.abs(f0) - 1))
    
    def err(self,params,f,s21):
        # f0, Qi, Qe, phi = params
        y = np.abs(s21) - np.abs(self.invS21(f, params) )
        return np.sum(np.abs(y)**2)

    def fitCavity(self,x,y):
        f, s = self.deductPhase(x,y)
        s = s[0]/np.max(np.abs(s[0]))
        f0, Qi, Qe, phi, QL = self.guessParams(f,s)
        res = bh(self.err,(f0, Qi, Qe, phi),niter = 100,\
            minimizer_kwargs={"method":"Nelder-Mead","args":(f, 1/s)}) 
        f0, Qi, Qe, phi = res.x
        QL = 1 / (1 / Qi + 1 / Qe)
        return res,self.invS21

def amp_opt(y,s,radio=0.5):
    s0=[]
    s1=[]
    for i in range(len(s)):
        s1 = (s[i]>(np.max(s[i])*radio+np.min(s[i])*(1-radio)))
        if i ==0:
            s0 = (s[i]>(np.max(s[i])*radio+np.min(s[i])*(1-radio))) 
        if list(s0)!=list(s1):
            for i,j in enumerate(s0):
                if j==False:
                    s1[i]=j
        s0=s1
    val = []
    indexs=[]
    for i,j in enumerate(s0):
        if j:
            val.append(y[i])
            indexs.append(i)
    indexw = np.argmax([np.mean(s[:,i]) for i in indexs])
    return y[indexs[indexw]]

    

def run_points(s0,si,idx=0,radio=0.5):
    s1 = (si>(np.max(si)*radio+np.min(si)*(1-radio)))
    if idx ==0:
        s0 = (si>(np.max(si)*radio+np.min(si)*(1-radio))) 
    if list(s0)!=list(s1):
        for n,j in enumerate(s1):
            if j==False:
                s0[n]=j
    return s0

  


# from qiskit import *
# from qiskit.quantum_info import Operator
# from qiskit.compiler import transpile
# from qiskit.compiler import transpiler

# from qiskit import QuantumCircuit, execute, Aer
# import numpy as np
# from numpy import pi
# from qiskit import Aer, transpile
# from qiskit.tools.visualization import plot_histogram, plot_state_city
# from qiskit.providers.aer.library import save_unitary
# import qiskit.quantum_info as qi

# from numpy import pi
# import numpy as np
# from collections.abc import Iterable
# def toQiskitcirc(circuit,qubits,show=False):
#     qubits = list(qubits)
#     qc = QuantumCircuit(len(qubits))
#     for c in circuit:
#     #     print(c)
#         if c[0] == 'CZ':
#             idx1, idx2 = qubits.index(c[1][0]), qubits.index(c[1][1])
#             qc.cz(idx1, idx2)
# #         elif 'rfUnitary' in c[0]:
# #             idx = qubits.index(c[1])
# #             qc.ry(c[0][1],idx)
#         elif 'Rz' in c[0]:
#             idx = qubits.index(c[1])
#             qc.rz(c[0][1],idx)
#         elif 'Ry' in c[0]:
#             idx = qubits.index(c[1])
#             qc.ry(c[0][1],idx)
#         elif 'Rx' in c[0]:
#             idx = qubits.index(c[1])
#             qc.rx(c[0][1],idx)
#         elif 'X/2' == c[0]:
#             idx = qubits.index(c[1])
#             qc.rx(np.pi/2,idx)
#         elif 'X' == c[0]:
#             idx = qubits.index(c[1])
#             qc.x(idx)
#         elif '-X/2' == c[0]:
#             idx = qubits.index(c[1])
#             qc.rx(-np.pi/2,idx)
#         elif 'H' == c[0]:
#             idx = qubits.index(c[1])
#             qc.h(idx)
#         elif 'Barrier' == c[0]:
#             qc.barrier()
#         elif 'delay' == c[0][0]:
#             idx = qubits.index(c[1])
#             qc.delay(c[0][1],idx)
#         elif 'Id' == c[0]:
#             idx = qubits.index(c[1])
#             qc.i(idx)
#         else:
#             pass
#     if show:
#         qc.draw(output='mpl')
#     return qc

# def getCounts(qc,shots=30000):
#     qc.measure_all()
#     backend = Aer.get_backend('qasm_simulator')
#     job = execute(qc,backend,shots=shots,basis_gates=None)
#     result = job.result()
#     counts = result.get_counts()
#     return counts
        


import asyncio, inspect
import functools

import nest_asyncio
import numpy as np
from scipy.optimize import minimize
from bayes_opt import BayesianOptimization as bayes
from bayes_opt.logger import JSONLogger
from bayes_opt.event import Events
from bayes_opt.util import load_logs

nest_asyncio.apply()


def get_initial_simplex(start, senstive=None):
    if senstive is None:
        senstive = np.ones(len(start))

    initial_simplex = [list(start)]
    for i, v in enumerate(senstive):
        row = list(start)
        row[i] += v
        initial_simplex.append(row)
    return np.asarray(initial_simplex)


def parameter_filter(x, dec=None, high=None, low=None):
    if dec is not None:
        x = np.asarray([np.round(a, decimals=d) for a, d in zip(x, dec)])
    if high is None and low is None:
        return x
    return x.clip(low, high)


def optimize(target,
             start,
             senstive=None,
             dec=None,
             high=None,
             low=None,
             method='Nelder-Mead',
             print_info=False,
             tol = None,
             kws={},
             options={}):
    """
    搜索最小值
    target: 目标函数
    start: 起始点
    senstive: list 敏感度
    dec: list 小数位数
    high: list 上限
    low: list 下限
    print_info: bool
    """

    optimizedTargetValue = None

    if asyncio.iscoroutinefunction(target):

        def _target(*x):
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(target(*x,**kws))
    else:
        _target = target

    @functools.lru_cache()
    def cache_target(*x):
        return _target(*x,**kws)

    def f(x):
        nonlocal optimizedTargetValue
        x = parameter_filter(x, dec, high, low)
        ret = cache_target(*x)
        if print_info:
            print('.', end='')
        if optimizedTargetValue is None or optimizedTargetValue > ret:
            optimizedTargetValue = ret
            if print_info:
                print('o')
                print(list(x), ret, end='   ')
        return ret
    options['initial_simplex'] = get_initial_simplex(start, senstive)

    ret = minimize(
        f,
        start,
        method=method,
        tol=tol,
        options=options)

    if print_info:
        print('\n', cache_target.cache_info())

    ret['x'] = parameter_filter(ret['x'], dec, high, low)

    return ret

def optimize_bayes(target,
             start,
             senstive=None,
             dec=None,
             high=None,
             low=None,
             print_info=False,
             init_points=5,
             n_iter=25,
             saveStatus=False,
             loadStatus=False,
             path=r"D:\skzhao\bayes_optimize_paras\bayes_paras.json"):
    """
    搜索最大值
    target: 目标函数
    start: 起始点
    senstive: list 敏感度
    dec: list 小数位数
    high: list 上限
    low: list 下限
    print_info: bool
    n_iter: 优化迭代次数
    """
    verbose= 2 if print_info else 1
    insp = inspect.getfullargspec(target)
    kwsLst = insp[0]

    pbounds, params_start = {}, {}
    for i, j in enumerate(kwsLst):
        pbounds[j] = (low[i],high[i])
        params_start[j] = start[i]

    # print('pbounds:',pbounds,'params_start:',params_start)

    if asyncio.iscoroutinefunction(target):

        def _target(**x):
            loop = asyncio.get_event_loop()
            return loop.run_until_complete(target(**x))
    else:
        _target = target

    @functools.lru_cache()
    def cache_target(**x):
        return _target(**x)

    rf_bo = bayes(cache_target,pbounds=pbounds,verbose=verbose,random_state=7)

    if loadStatus:
        load_logs(rf_bo, logs=[path])

    rf_bo.probe(params=params_start,lazy=True,)

    rf_bo.maximize(init_points=init_points,n_iter=n_iter,allow_duplicate_points=True)

    if saveStatus:
        logger = JSONLogger(path=path)
        rf_bo.subscribe(Events.OPTIMIZATION_STEP, logger)

    return rf_bo.max, rf_bo


import functools
import skopt, pickle
import numpy as np
from tqdm.notebook import trange

def parameter_filter(x, dec=None, high=None, low=None):
    if dec is not None:
        x = np.asarray([np.round(a, decimals=d) for a, d in zip(x, dec)])
    if high is None and low is None:
        return x
    return x.clip(low, high)

def minimize_bayes(target,
             start,
             senstive=None,
             dec=None,
             high=None,
             low=None,
             kws = {},
             print_info=False,
             n_iter=25,
             n_jobs = 10,
             model_queue_size = 1,
             initOptclass = None,
             peak = None,
             pbar = True):
    
#     @functools.lru_cache()
#     def cache_target(*x):
#         return target(*x,**kws)
    low = np.asarray([np.round(float(a), decimals=d) for a, d in zip(low, dec)])
    high = np.asarray([np.round(float(a), decimals=d) for a, d in zip(high, dec)])
    
    if initOptclass is None:
        opt = skopt.Optimizer(list(zip(low,high)),n_jobs=n_jobs,model_queue_size=model_queue_size) 
    else:
        opt = initOptclass
    
    y = target(start,**kws)
    ymin = y
    opt.tell(list(start),y)
    print(f'x0={list(start)},y0={y}')

    scan_iter = trange(n_iter,desc=f'Optimizing {target.__name__}') if pbar and not print_info else range(n_iter)
    
    for i in scan_iter:
        x = opt.ask()
        x = parameter_filter(x, dec, high, low)
        y = target(x,**kws)
        opt.tell(list(x),y)
        if print_info:
            if y > ymin:
                print(f'step{i+1},  x={list(x)},  y={y}')
            else:
                ymin = y
                pp_info = f'step{i+1},  x={list(x)},  y={y}'
                print(f"\033[1;30;41m{pp_info}\033[0m")
        
        if peak is not None and y <= peak:
            return opt
        
    return opt


import numpy as np
import matplotlib.pyplot as plt
from datetime import date
from pathlib import Path
import os


def create_filename(data={},title = 'title',fold=None,basepath=r'G:\QuLabData\skzhao\data'):
    today=date.today()
    basepath = Path(basepath) if fold is None else Path(basepath+f'\{fold}')
    basepath = basepath/today.strftime('%Y%m%d')
    foder=os.path.exists(basepath)
    # if not foder:
    #     os.makedirs(basepath)
    #     np.savez(basepath/r'filenames.npz',filenames={})
    #     print('New foder!')
    # else:
    #     print('Exist!')
    if foder:
        filenames = np.load(basepath/r'filenames.npz',allow_pickle=True)['filenames'].tolist()
        print('Exist!')
    else:
        os.makedirs(basepath)
        filenames = {}
        np.savez(basepath/r'filenames.npz',filenames=filenames)
        print('New folder!')

    # a = 0
    # title_new = title
    # while basepath / (title_new+'.npz') in [f for f in basepath.iterdir()]:
    #     a +=1
    #     title_new = title+f'_{a}'

    a = filenames[title]  if title in filenames else 0
    title_new = title+f'_{a}'
    filenames[title]  = a+1
    np.savez(basepath/r'filenames.npz',filenames=filenames)

    np.savez(basepath / title_new, **data)

    return basepath / title_new

def saveconfig_old(data={}, title = 'title', fig=None, fold=None, close=False, basepath=r'G:\QuLabData\skzhao\data'):

    filename = create_filename(data=data,title=title,fold=fold,basepath=basepath)

    if fig is not None:
        fig.savefig(filename)
    if close:
        plt.close()
    return filename

def saveconfig(task_id, title = 'title', fig=None, fold=None, show=True, basepath=r'G:\QuLabData\skzhao\data',cover=False):
    from datetime import date
    from pathlib import Path
    import os
    today=date.today()
    basepath = Path(basepath) if fold is None else Path(basepath+f'\{fold}')
    basepath = basepath/today.strftime('%Y%m%d')
    foder=os.path.exists(basepath)
    if foder:
        filenames = np.load(basepath/r'filenames.npz',allow_pickle=True)['filenames'].tolist()
        print('Exist!')
    else:
        os.makedirs(basepath)
        filenames = {}
        np.savez(basepath/r'filenames.npz',filenames=filenames)
        print('New folder!')
        
        
    a = filenames[title]  if title in filenames else 0
    title_new = title+f'_{a}'
    filenames[title]  = a+1
    np.savez(basepath/r'filenames.npz',filenames=filenames)


    # filename = create_filename(data=data,title=title,fold=fold,basepath=basepath)

    if fig is not None:
        name = fig.layout.title.text
        if '_id' not in name:
            name += f'_id={task_id}'
            fig.update_layout(title=name)
        title_new_html = title_new + '.html'
        title_new_fig = title_new + '.png'
        fig.write_html(basepath / title_new_html)
        # fig.write_image(basepath / title_new_fig,engine='kaleido')
    if show:
        fig.show()
    return basepath / title_new_html


class common():
    def __init__(self,**kws):
        attribute = ['qubits_read','all_qubits','all_couplers','all_jpas','ampLst_read','biasLst_read','biasLst_qubit',\
            'biasLst_couplers','biasLst_jpa','ampLst_jpa','output_jpa','phaseLst_jpa','biasLst_reset','duringLst_reset',
            'edgeLst_reset','ring_up_amp','ring_up_time']
        for j in attribute:
            self.__setattr__(j,None)
        if len(kws) != 0:
            for i in kws:
                self.__setattr__(i,kws[i])

        # self.kernel = kernel
        self.readmatrix = {}
        self.readoutM = {}
        self.info = {}

    def asdict(self):
        arg = self.__dict__
        arg_copy = arg.copy()
        # del arg_copy['kernel']
        return arg_copy

    def check_repeat(self):
        assert len(set(self.all_couplers)) == len(self.all_couplers)
        assert len(set(self.all_qubits)) == len(self.all_qubits)
        assert len(set(self.qubits_read)) == len(self.qubits_read)
        assert len(set(self.all_jpas)) == len(self.all_jpas)

    def replace(self,**kws):
        for i in kws:
            assert hasattr(self,i)
            self.__setattr__(i,kws[i])

    def getvalue(self,name):
        return self.__getattribute__(name)

    def setvalue(self,name,value):
        # argLst = self.__getattribute__(name)
        # idx = argLst.index()
        self.__setattr__(name,value)

    def initCfg(self,kernel):
        self.check_repeat()
        biasLst_read = [kernel.get(f'gate.Measure.{q}.params.bias') for q in self.qubits_read]
        ampLst_read = [kernel.get(f'gate.Measure.{q}.params.amp') for q in self.qubits_read]
        ring_up_amp = [kernel.get(f'gate.Measure.{q}.params.ring_up_amp') for q in self.qubits_read]
        ring_up_time = [kernel.get(f'gate.Measure.{q}.params.ring_up_time') for q in self.qubits_read]
        shelving = [kernel.get(f'gate.Measure.{q}.params.shelving') for q in self.qubits_read]

        biasLst_jpa = [kernel.get(f'{q}.setting.OFFSET') for q in self.all_jpas]
        output_jpa = [kernel.get(f'{q}.setting.OUTPUT') for q in self.all_jpas]
        ampLst_jpa = [kernel.get(f'gate.Amplify.{q}.params.amp') for q in self.all_jpas]
        phaseLst_jpa = [kernel.get(f'gate.Amplify.{q}.params.phase') for q in self.all_jpas]

        biasLst_reset = [kernel.get(f'gate.Reset.{q}.params.amp') for q in self.all_qubits]
        duringLst_reset = [kernel.get(f'gate.Reset.{q}.params.duration') for q in self.all_qubits]
        edgeLst_reset = [kernel.get(f'gate.Reset.{q}.params.edge') for q in self.all_qubits]

        biasLst_couplers = [kernel.get(f'{coupler0}.bias') for coupler0 in self.all_couplers]
        biasLst_qubit = [kernel.get(f'{q}.bias') for q in self.all_qubits]
        self.biasLst_read = biasLst_read
        self.ampLst_jpa = ampLst_jpa
        self.biasLst_jpa = biasLst_jpa
        self.output_jpa = output_jpa
        self.phaseLst_jpa = phaseLst_jpa
        self.ampLst_read = ampLst_read
        self.ring_up_amp = ring_up_amp
        self.ring_up_time = ring_up_time
        self.biasLst_couplers = biasLst_couplers
        self.biasLst_qubit = biasLst_qubit
        self.shelving = shelving
        self.biasLst_reset = biasLst_reset
        self.duringLst_reset = duringLst_reset
        self.edgeLst_reset = edgeLst_reset

        print('biasLst_read=',biasLst_read)
        print('ampLst_read=',ampLst_read)
        print('biasLst_jpa=',biasLst_jpa)
        print('ampLst_jpa=',ampLst_jpa)
        print('output_jpa=',output_jpa)
        print('phaseLst_jpa=',phaseLst_jpa)
        print(f'shelving={self.shelving}')
        print('ring_up_amp=',ring_up_amp)
        print('ring_up_time=',ring_up_time)
        print('biasLst_couplers=',biasLst_couplers)
        print('biasLst_qubit=',biasLst_qubit)
        print('biasLst_reset=',biasLst_reset)
        print('duringLst_reset=',duringLst_reset)
        print('edgeLst_reset=',edgeLst_reset)

    def write(self, kernel, name = None):
        # kernel = self.kernel if kernel is None else kernel
        if name is None:
            # count = 0
            for i,qubit in enumerate(self.all_qubits):
                kernel.set(f'{qubit}.bias',self.biasLst_qubit[i])
                kernel.set(f'gate.Reset.{qubit}.params.amp',self.biasLst_reset[i])
                kernel.set(f'gate.Reset.{qubit}.params.duration',self.duringLst_reset[i])
                kernel.set(f'gate.Reset.{qubit}.params.edge',self.edgeLst_reset[i])
                if qubit in self.qubits_read:
                    idx_meas = self.qubits_read.index(qubit)
                    kernel.set(f'gate.Measure.{qubit}.params.bias',self.biasLst_read[idx_meas])
                    kernel.set(f'gate.Measure.{qubit}.params.amp',self.ampLst_read[idx_meas])
                    kernel.set(f'gate.Measure.{qubit}.params.ring_up_amp',self.ring_up_amp[idx_meas])
                    kernel.set(f'gate.Measure.{qubit}.params.ring_up_time',self.ring_up_time[idx_meas])
                    # count += 1
                    
            for i, coupler in enumerate(self.all_couplers):
                kernel.set(f'{coupler}.bias',self.biasLst_couplers[i])
        else:
            for i,qubit in enumerate(self.qubits_read):
                paras = self.biasLst_read if name == 'bias' else self.ampLst_read
                kernel.set(f'gate.Measure.{qubit}.params.{name}',paras[i])

    def loadcfg(self,kernel):
        cfg = kernel.get('measure_mrw')
        for k in cfg:
            self.__setattr__(k,cfg[k])

        v = self.asdict()
        for k in v:
            print(k,v[k])

    def uploadcfg(self,kernel):

        v = self.asdict()
        for k in v:
            kernel.set(f'measure_mrw.{k}',v[k])

def readPop(result,readmatrix=False,reshape=True,datatype='population',
            statenum=2,qubits_read=None,caliMatrix=False,method='count1'):
    
    from collections import Counter
    from itertools import product
    result = result if isinstance(result,dict) else result()
    measure = result['meta']['arguments']['measure_mrw']
    readmatrix = measure['readmatrix'] if readmatrix else None
#     info = measure.info
    qubits_read = measure['qubits_read'] if qubits_read is None else qubits_read
    idxLst = np.array([measure['qubits_read'].index(q) for q in qubits_read])
    arg = list(result['index'].keys())
    
    def statePop(data=None,qubits_read_=None,reshape=reshape,caliMatrix=caliMatrix):
        qubits_read_, idxLst_ = (qubits_read,idxLst) if qubits_read_ is None \
        else (qubits_read_,np.array([measure.qubits_read.index(q) for q in qubits_read_]))

        coQubit = ''.join((*qubits_read,))
        caliMatrix = measure['readoutM'][coQubit] if caliMatrix else None
        
        def count1(data,qubits_read,statenum=2):

            state = product(2**np.arange(statenum), repeat=len(qubits_read))
            c = Counter(map(tuple, data))
            pop = [c[tuple(i)] for i in state]
            return np.array(pop)/np.sum(pop)
        def count2(data,qubits_read,statenum=2):
            state = product(2**np.arange(statenum), repeat=len(qubits_read))
            # print(list(state),data.shape)
            c = Counter(map(tuple, data))
            # print(c)
            pop = [c[tuple(i)] for i in state]
            return np.array(pop)/np.sum(pop)
        methodLst = {'count1':count1,'count2':count2}
        count_method = methodLst[method]
        s = result['state']
        y = s[:,:,idxLst_] if data is None else data
#         y = s
        # print(y.shape,y[0,:,:].shape)
        repeats = np.shape(s)[-2]
        pop = np.array([count_method(y[i,:,:],qubits_read,statenum) for i in range(np.shape(s)[0])])
        # pop = np.array([count1(y[i,:,:],qubits_read,statenum) for i in range(np.shape(y)[0])])

        coMatrix = np.mat(caliMatrix) if caliMatrix is not None else np.mat(np.eye(statenum**len(qubits_read)))
        pop_cali = np.array((coMatrix.I*np.mat(pop).T).T).flatten()
        # print(pop_cali)
        if reshape:
            length = []
            for idx_arg in arg:
                length.append(len(result['index'][idx_arg]))
            length.append(statenum**len(qubits_read))
            pop_cali = np.array(pop_cali).reshape(*length)
        return pop_cali
            
    def population():
        
        p1 = result['population']
        p0 = 1-p1
        if readmatrix is None:
            pop_cali = p1
        else:
            pop_cali = []
            for i,q in enumerate(qubits_read):
                mat = readmatrix[q] if q in readmatrix else np.mat(np.eye(2))
                pop = np.array(np.mat(mat).I*np.mat([p0[:,i],p1[:,i]]))
                pop_cali.append(pop[1,:])
            pop_cali = np.array(pop_cali).T
            
        if reshape:
            length = []
            for idx_arg in arg:
                length.append(len(result['index'][idx_arg]))
            length.append(len(qubits_read))
            pop_cali = (pop_cali).reshape(*length)
        return pop_cali

    def state2Pop(pop):
        state = product(range(statenum), repeat=len(qubits_read))
        state = np.array(list(state))
        pop_cali = []
        for i,q in enumerate(qubits_read):
            idx = state[:,i]==1
            p1 = pop[...,idx]
            p1 = np.sum(p1,axis=-1)
            pop_cali.append(p1)
        pop_cali = np.array(pop_cali).T

        return pop_cali
    
    if datatype == 'population':
        pop_cali = population()
        
    if datatype == 'state':
        pop_cali = statePop()

    if datatype == 'state2Pop':
        pop_cali = statePop(reshape=True)
        pop_cali = state2Pop(pop_cali)
        
    return pop_cali



import setuptools
import os
from distutils.command.register import register as register_orig
from distutils.command.upload import upload as upload_orig

from setuptools import setup


class register(register_orig):

    def _get_rc_file(self):
        return os.path.join('.', '.pypirc')

class upload(upload_orig):

    def _get_rc_file(self):
        return os.path.join('.', '.pypirc')


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="qufit",
    version="1.1.69",
    author="赵寿宽(sk zhao)",
    author_email="2396776980@qq.com",
    description="本模块包括了超导量子计算实验中常用模型的拟合函数于数据处理方法。",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://pypi.org/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
    ],
    cmdclass={
        'register': register,
        'upload': upload,
    },
    # install_requires=[
    #     'qufit',
    #     'numpy',
    #     'pandas',
    #     'scikit-learn',
    #     'scipy',
    #     'sympy',
    #     'matplotlib',
    #     'imp',
    #     'math',
    #     'bayesian-optimization',
    #     'nest_asyncio',
    #     'scikit-optimizer',
    #     'pickle',
    #     'plotly',
        
    # ],
)





import socket, pickle, time
class clientDriver():
    
    def __init__(self, addr=None, port=6969):
#         socket.setdefaulttimeout(100000)
        self.addr = addr
        self.port = port
        
    def connect(self):
        
        client = socket.socket()
        client.settimeout(1000000)
        client.connect((self.addr,self.port))
        print('connected',client)
        self.handle = client
        
    def send_data(self,counts_dict):
        counts_dict = pickle.dumps(counts_dict)
#         print('send ready?')
        self.handle.send(pickle.dumps({'0':'ready?','1':len(counts_dict)}))
        data = self.handle.recv(1024*5000)
        data = pickle.loads(data)
#         print('recv ready!')
        
        if data == 'ready!':
            self.handle.send(counts_dict)
    
    def recv_data(self):
        
        data = self.handle.recv(1024*5000)
        data = pickle.loads(data)
#         print('recv ready?')
        
        if data['0'] == 'ready?':
            length = data['1']
            self.handle.send(pickle.dumps('ready!'))
            counts_dict = b''
            while 1:
    #             print('send ready!')
                counts_dict += self.handle.recv(1024*5000)
                if len(counts_dict) == length:
                    break
            counts_dict = pickle.loads(counts_dict)
        else:
            pass
            
        return counts_dict
    
    def close(self):
        self.handle.close()   
        
        
        
import socket, pickle, time
class serverDriver():
    
    def __init__(self, addr=None, user=1, port=6969):
#         socket.setdefaulttimeout(100000)
        self.addr = addr
        self.user = user
        self.port = port
        
    def mount(self):
        
        server = socket.socket()
        server.settimeout(10000000)
        server.bind((self.addr,self.port))
        server.listen(self.user)
        print('waiting the call')
        conn, addr = server.accept()
        self.handle = conn
        
    def send_data(self,counts_dict):
        counts_dict = pickle.dumps(counts_dict)
#         print('send ready?')
        self.handle.send(pickle.dumps({'0':'ready?','1':len(counts_dict)}))
        data = self.handle.recv(1024*5000)
        data = pickle.loads(data)
#         print('recv ready!')
        
        if data == 'ready!':
            self.handle.send(counts_dict)
    
    def recv_data(self):
        
        data = self.handle.recv(1024*5000)
        data = pickle.loads(data)
#         print('recv ready?')
        
        if data['0'] == 'ready?':
            length = data['1']
            self.handle.send(pickle.dumps('ready!'))
            counts_dict = b''
            while 1:
    #             print('send ready!')
                counts_dict += self.handle.recv(1024*5000)
                if len(counts_dict) == length:
                    break
            counts_dict = pickle.loads(counts_dict)
        else:
            pass
            
        return counts_dict
    
    def close(self):
        self.send_data('close')
        self.handle.close()